"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging

from cfn_policy_validator.parsers.resource.kms import KmsKeyPolicyParser
from cfn_policy_validator.parsers.resource.s3 import S3BucketPolicyParser, S3AccessPointPolicyParser, \
    S3MultiRegionAccessPointPolicyParser, S3BucketAclParser
from cfn_policy_validator.parsers.resource.sns import SnsTopicPolicyParser
from cfn_policy_validator.parsers.resource.sqs import SqsQueuePolicyParser
from cfn_policy_validator.parsers.resource.lambda_aws import LambdaPermissionPolicyParser, LambdaLayerVersionPermissionParser
from cfn_policy_validator.parsers.resource.secrets_manager import SecretsManagerPolicyParser

from cfn_policy_validator.parsers.utils.topological_sorter import TopologicalSorter


LOGGER = logging.getLogger("cfn-policy-validator")


class ResourceParser:
    """
    Passes parsing for resource-based policies to resource-specific parsers.  To add a new parser, modify the parsers
    dictionary below.
    """

    @classmethod
    def parse(cls, template, account_config):
        # topologically sort which allows us to process dependent resources first
        sorter = TopologicalSorter(template)
        sorted_resources = sorter.sort_resources()

        parsers = {
            'AWS::S3::AccessPoint': S3AccessPointPolicyParser(),
            'AWS::S3::MultiRegionAccessPointPolicy': S3MultiRegionAccessPointPolicyParser(),
            'AWS::S3::Bucket': S3BucketAclParser(),
            'AWS::S3::BucketPolicy': S3BucketPolicyParser(),
            'AWS::SQS::QueuePolicy': SqsQueuePolicyParser(),
            'AWS::SNS::TopicPolicy': SnsTopicPolicyParser(),
            'AWS::KMS::Key': KmsKeyPolicyParser(),
            'AWS::Lambda::Permission': LambdaPermissionPolicyParser(account_config),
            'AWS::Lambda::LayerVersionPermission': LambdaLayerVersionPermissionParser(account_config.partition),
            'AWS::SecretsManager::ResourcePolicy': SecretsManagerPolicyParser()
        }

        invoked_parsers = set()
        for resource in sorted_resources:
            resource_type = resource.value['Type']
            parser = parsers.get(resource_type)
            if parser is not None:
                LOGGER.info(f'Parsing resource type {resource_type} with logical name {resource.logical_name}..')
                parser.parse(resource.logical_name, resource.value)
                invoked_parsers.add(parser)

        # If multiple cfn resource types exist in the same template that should be validated together, combine them
        # into a single resource as it may be confusing to validate the same resource twice.
        # an example is if an AWS::S3::Bucket is defined with ACLs and also has an AWS::S3::BucketPolicy.
        # If they are in different templates, it's OK to evaluate them separately
        for parser in invoked_parsers:
            merge_policies = getattr(parser, "merge_policies", None)
            if merge_policies is not None:
                parser.merge_policies(invoked_parsers)

        return [policy for parser in invoked_parsers for policy in parser.get_policies()]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging

from abc import ABC

from botocore.exceptions import ClientError

from cfn_policy_validator import client
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.parsers.identity_schemas import groups_schema, managed_policy_schema, \
	roles_schema, inline_policy_schema, users_schema, permission_set_schema
from cfn_policy_validator.parsers.utils.arn_generator import ArnGenerator
from cfn_policy_validator.parsers.output import Role, User, Policy, Group, PermissionSet
from cfn_policy_validator.parsers.utils.topological_sorter import TopologicalSorter


LOGGER = logging.getLogger("cfn-policy-validator")


class IdentityParser:
	"""
	Parses identity policies from IAM CloudFormation resources.  Note that identity policies may have dependencies
	on each other (e.g. policies attached to principals).
	"""

	@classmethod
	def parse(cls, template, account_config):
		parsers = {
			'AWS::IAM::Role': RoleParser(account_config.region),
			'AWS::IAM::Policy': InlinePolicyParser(),
			'AWS::IAM::ManagedPolicy': ManagedPolicyParser(account_config),
			'AWS::IAM::User': UserParser(account_config.region),
			'AWS::IAM::Group': GroupParser(account_config.region),
			'AWS::SSO::PermissionSet': PermissionSetParser(account_config.region)
		}

		# topologically sort which allows us to process dependent resources first
		sorter = TopologicalSorter(template)
		sorted_resources = sorter.sort_resources()

		for resource in sorted_resources:
			resource_type = resource.value['Type']
			parser = parsers.get(resource_type)
			if parser is not None:
				LOGGER.info(f'Parsing resource type {resource_type} with logical name {resource.logical_name}..')
				parser.parse(resource.logical_name, resource.value)

		orphaned_policies = cls.get_orphaned_policies()

		return list(RoleParser.roles.values()), \
			list(UserParser.users.values()), \
			list(GroupParser.groups.values()), \
			list(PermissionSetParser.permission_sets.values()), \
			orphaned_policies

	@classmethod
	def get_orphaned_policies(cls):
		"""
		Orphaned policies are policies that are not attached to any principal
		"""
		role_policies = [policy for role in RoleParser.roles.values() for policy in role.Policies]
		user_policies = [policy for user in UserParser.users.values() for policy in user.Policies]
		group_policies = [policy for group in GroupParser.groups.values() for policy in group.Policies]
		permission_set_policies = [policy for permission_set in PermissionSetParser.permission_sets.values()
								   for policy in permission_set.Policies]

		all_attached_policies = list(set(role_policies) | set(user_policies) | set(group_policies) | set(permission_set_policies))
		all_managed_policies = list(ManagedPolicyParser.managed_policies.values())
		all_inline_policies = list(InlinePolicyParser.inline_policies.values())

		orphaned_policies = [managed_policy for managed_policy in all_managed_policies if managed_policy not in all_attached_policies]
		orphaned_policies.extend([inline_policy for inline_policy in all_inline_policies if inline_policy not in all_attached_policies])

		return orphaned_policies


class AttachedPolicyParser:
	"""
	Handles parsing managed and inline policies from other resources
	"""
	def __init__(self, region):
		self.client = client.build('iam', region)

	@staticmethod
	def parse_inline_policies(principal, properties):
		inline_policies = properties.get('Policies', [])
		for inline_policy in inline_policies:
			policy_name = inline_policy['PolicyName']
			policy_document = inline_policy['PolicyDocument']

			policy = Policy(policy_name, policy_document)
			principal.add_policy(policy)

	def parse_managed_policies(self, principal, properties, property_name='ManagedPolicyArns'):
		managed_policy_arns = properties.get(property_name, [])
		for arn in managed_policy_arns:
			policy = ManagedPolicyParser.managed_policies.get(arn)
			if policy is not None:
				principal.add_policy(policy)
				continue

			try:
				# if the ARN is not a managed policy in the template, pull it from the environment
				response = self.client.get_policy(PolicyArn=arn)
			except ClientError as e:
				if e.response['Error']['Code'] == 'NoSuchEntity':
					raise ApplicationError(f'Could not find managed policy with {arn} in template or in environment.')
				else:
					raise

			policy_name = response['Policy']['PolicyName']
			default_version_id = response['Policy']['DefaultVersionId']
			policy_path = response['Policy']['Path']

			response = self.client.get_policy_version(PolicyArn=arn, VersionId=default_version_id)
			policy_document = response['PolicyVersion']['Document']

			# it must be a valid ARN if we made it this far
			arn_parts = arn.split(':')
			account_id = arn_parts[4]
			# record whether this is an AWS managed policy because we may treat them slightly different for validation
			is_aws_managed_policy = account_id.lower() == 'aws'

			policy = Policy(policy_name, policy_document, policy_path, is_aws_managed_policy)
			principal.add_policy(policy)


class RoleParser:
	""" Parser for AWS::IAM::Role
	"""

	# store roles for the lifetime of the run for static lookup by other parsers
	roles = {}

	def __init__(self, region):
		self.attached_policy_parser = AttachedPolicyParser(region)
		RoleParser.roles = {}

	@classmethod
	def get_role_by(cls, name):
		for role in cls.roles.values():
			if role.RoleName == name:
				return role

		# this means that it the template is referencing a role external to the template,
		# but we still want to validate policies in the template
		return None

	def parse(self, resource_name, resource):
		evaluated_resource = resource.eval(roles_schema)

		properties = evaluated_resource['Properties']

		path = properties.get('Path', '/')
		trust_policy = properties['AssumeRolePolicyDocument']

		# if RoleName is not a property, use the resource name
		role_name = properties.get('RoleName', resource_name)

		role = Role(role_name, path, trust_policy)

		self.attached_policy_parser.parse_inline_policies(role, properties)
		self.attached_policy_parser.parse_managed_policies(role, properties)

		self.roles[resource_name] = role


class UserParser:
	""" AWS::IAM::User
	"""

	# store users for the lifetime of the run for static lookup by other parsers
	users = {}

	def __init__(self, region):
		self.attached_policy_parser = AttachedPolicyParser(region)
		UserParser.users = {}

	@classmethod
	def get_user_by(cls, name):
		for user in cls.users.values():
			if user.UserName == name:
				return user

		# this means that it the template is referencing a role external to the template,
		# but we still want to validate policies in the template
		return None

	def parse(self, resource_name, resource):
		evaluated_resource = resource.eval(users_schema)

		properties = evaluated_resource.get('Properties', {})

		path = properties.get('Path', '/')
		user_name = properties.get('UserName', resource_name)

		user = User(user_name, path)

		self.attached_policy_parser.parse_inline_policies(user, properties)
		self.attached_policy_parser.parse_managed_policies(user, properties)

		self.users[resource_name] = user


class GroupParser:
	""" AWS::IAM::Group
	"""

	# store groups for the lifetime of the run for static lookup by other parsers
	groups = {}

	def __init__(self, region):
		self.attached_policy_parser = AttachedPolicyParser(region)
		GroupParser.groups = {}

	@classmethod
	def get_group_by(cls, name):
		for group in cls.groups.values():
			if group.GroupName == name:
				return group

		# this means that the template is referencing a role external to the template,
		# but we still want to validate policies in the template
		return None

	def parse(self, resource_name, resource):
		evaluated_resource = resource.eval(groups_schema)
		# there are no required properties for a group, so a group without properties is OK
		properties = evaluated_resource.get('Properties', {})

		path = properties.get('Path', '/')
		group_name = properties.get('GroupName', resource_name)

		group = Group(group_name, path)

		self.attached_policy_parser.parse_inline_policies(group, properties)
		self.attached_policy_parser.parse_managed_policies(group, properties)

		self.groups[resource_name] = group


class PermissionSetParser:
	""" Parser for AWS::SSO::PermissionSet
	"""

	permission_sets = {}

	def __init__(self, region):
		self.attached_policy_parser = AttachedPolicyParser(region)
		PermissionSetParser.permission_sets = {}

	def parse(self, resource_name, resource):
		evaluated_resource = resource.eval(permission_set_schema)

		properties = evaluated_resource['Properties']

		name = properties['Name']

		permission_set = PermissionSet(name)

		self.parse_inline_policy(permission_set, properties)
		self.attached_policy_parser.parse_managed_policies(permission_set, properties, property_name="ManagedPolicies")

		self.permission_sets[resource_name] = permission_set

	@staticmethod
	def parse_inline_policy(permission_set, properties):
		""" PermissionSets can only have a single inline policy so the parsing logic is slightly different
		"""
		inline_policy = properties.get('InlinePolicy')
		if inline_policy is None:
			return

		policy = Policy('InlinePolicy', inline_policy)
		permission_set.add_policy(policy)


class PolicyParser(ABC):
	"""
	Base class for common policy (inline and managed) parsing
	"""

	@staticmethod
	def parse_roles(policy, properties):
		roles_to_apply_policy = properties.get('Roles', [])
		for role_name in roles_to_apply_policy:
			referenced_role = RoleParser.get_role_by(role_name)
			if referenced_role is not None:
				referenced_role.add_policy(policy)

	@staticmethod
	def parse_users(policy, properties):
		users_to_apply_policy = properties.get('Users', [])
		for user_name in users_to_apply_policy:
			referenced_user = UserParser.get_user_by(user_name)
			if referenced_user is not None:
				referenced_user.add_policy(policy)

	@staticmethod
	def parse_groups(policy, properties):
		groups_to_apply_policy = properties.get('Groups', [])
		for group_name in groups_to_apply_policy:
			referenced_group = GroupParser.get_group_by(group_name)
			if referenced_group is not None:
				referenced_group.add_policy(policy)


class InlinePolicyParser(PolicyParser):
	""" AWS::IAM::Policy
	"""
	inline_policies = {}

	def __init__(self):
		super(InlinePolicyParser, self).__init__()
		InlinePolicyParser.inline_policies = {}

	def parse(self, _, resource):
		evaluated_resource = resource.eval(inline_policy_schema)

		properties = evaluated_resource['Properties']

		policy_name = properties['PolicyName']
		policy_document = properties['PolicyDocument']

		policy = Policy(policy_name, policy_document)

		self.parse_roles(policy, properties)
		self.parse_users(policy, properties)
		self.parse_groups(policy, properties)

		self.inline_policies[policy_name] = policy


class ManagedPolicyParser(PolicyParser):
	""" AWS::IAM::ManagedPolicy
	"""

	managed_policies = {}

	def __init__(self, account_config):
		super(ManagedPolicyParser, self).__init__()
		self.arn_generator = ArnGenerator(account_config)
		ManagedPolicyParser.managed_policies = {}

	def parse(self, resource_name, resource):
		evaluated_resource = resource.eval(managed_policy_schema)

		properties = evaluated_resource['Properties']

		policy_name = properties.get('ManagedPolicyName', resource_name)
		policy_document = properties['PolicyDocument']
		path = properties.get('Path', '/')

		policy = Policy(policy_name, policy_document, path)

		# attempt to see if this managed policy is attached to any roles, users or groups in the template
		self.parse_roles(policy, properties)
		self.parse_users(policy, properties)
		self.parse_groups(policy, properties)

		# generate an ARN to uniquely reference this managed policy since managed policy ARNs can also be referenced
		# from the principal side
		arn = self.arn_generator.try_generate_arn(policy_name, resource, 'Ref')
		self.managed_policies[arn] = policy



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.parsers.output import Policy, Resource


class SnsTopicPolicyParser:
    """ AWS::SNS::TopicPolicy
    """
    def __init__(self):
        self.topic_policies = []

    def parse(self, _, resource):
        evaluated_resource = resource.eval(sns_topic_policy_schema)
        properties = evaluated_resource['Properties']

        topic_arns = properties['Topics']
        policy_document = properties['PolicyDocument']

        for topic_arn in topic_arns:
            topic_name = topic_arn.split(':')[-1]

            policy = Policy('TopicPolicy', policy_document)
            resource = Resource(topic_name, 'AWS::SNS::Topic', policy)

            self.topic_policies.append(resource)

    def get_policies(self):
        return self.topic_policies


sns_topic_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'PolicyDocument': {
                    'type': 'object'
                },
                'Topics': {
                    'type': 'array',
                    'minItems': 1,
                    'items': {
                        'type': 'string'
                    }
                }
            },
            'required': ['PolicyDocument', 'Topics']
        }
    },
    'required': ['Properties']
}


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.common_schema import string_schema
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema


class FindInMapEvaluator:
	def __init__(self, mappings, node_evaluator):
		self.node_evaluator = node_evaluator
		self.mappings = mappings

	def evaluate(self, value, visited_values):
		validate_schema(value, find_in_map_schema, 'Fn::FindInMap')

		# these hardcoded path names help with debugging if there are eval issues
		map_name = self.node_evaluator.eval_with_validation(value[0], string_schema, path='Fn::FindInMap.0', visited_values=visited_values)
		top_level_key = self.node_evaluator.eval_with_validation(value[1], string_schema, path='Fn::FindInMap.1', visited_values=visited_values)
		second_level_key = self.node_evaluator.eval_with_validation(value[2], string_schema, path='Fn::FindInMap.2', visited_values=visited_values)

		result = self.mappings.get(map_name, {}) \
								.get(top_level_key, {}) \
								.get(second_level_key)

		if result is None:
			raise ApplicationError(f'Fn::FindInMap lookup failed. Unable to find value in Mappings.  Value: {value}')

		return result


find_in_map_schema = {
	'type': 'array',
	'items': [
		{},
		{},
		{}
	],
	'additionalItems': False
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
string_schema = {
	'type': 'string'
}

array_of_strings_schema = {
	'type': 'array',
	'items': {
		'type': 'string'
	}
}


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import datetime
import json


class Output:
    """
    Parser output representation.  Populated by identity and resource parsers.
    """

    def __init__(self, account_config):
        self.Region = account_config.region
        self.Account = account_config.account_id
        self.Partition = account_config.partition
        self.Roles = []
        self.PermissionSets = []
        self.Users = []
        self.Groups = []
        self.Resources = []
        self.OrphanedPolicies = []

    def print(self):
        as_json_string = json.dumps(self, default=self.default_to_json, indent=4)
        print(as_json_string)

    def to_json(self):
        return json.loads(
            json.dumps(self, default=self.default_to_json, indent=4)
        )

    @classmethod
    def default_to_json(cls, value):
        if isinstance(value, datetime.date):
            return value.isoformat()

        # allow a class to specify custom JSON serialization
        custom_to_json = getattr(value, "custom_to_json", None)
        if custom_to_json is None:
            return value.__dict__
        else:
            return custom_to_json()

    def __eq__(self, other):
        if not isinstance(other, Output):
            return False

        # written this way to allow for debugging
        permission_sets_are_equal = sorted(self.PermissionSets) == sorted(other.PermissionSets)
        roles_are_equal = sorted(self.Roles) == sorted(other.Roles)
        users_are_equal = sorted(self.Users) == sorted(other.Users)
        groups_are_equal = sorted(self.Groups) == sorted(other.Groups)
        resources_are_equal = sorted(self.Resources) == sorted(other.Resources)
        policies_are_equal = sorted(self.OrphanedPolicies) == sorted(other.OrphanedPolicies)
        return roles_are_equal and users_are_equal and groups_are_equal and \
               resources_are_equal and policies_are_equal and permission_sets_are_equal

    def __hash__(self):
        return hash((self.Roles, self.Users, self.Groups, self.Resources, self.OrphanedPolicies))


class IdentityWithPolicies:
    def __init__(self):
        self.Policies = []

    def add_policy(self, policy):
        self.Policies.append(policy)

    def __eq__(self, other):
        if not isinstance(other, IdentityWithPolicies):
            return False

        return self.Policies == other.Policies

    def __hash__(self):
        return hash(self.Policies)


class User(IdentityWithPolicies):
    def __init__(self, user_name, user_path):
        self.UserName = user_name
        self.UserPath = user_path
        super(User, self).__init__()

    def __eq__(self, other):
        if not isinstance(other, User):
            return False

        return self.UserName == other.UserName and\
            self.UserPath == other.UserPath and\
            super().__eq__(other)

    def __lt__(self, other):
        return self.UserName < other.UserName

    def __hash__(self):
        return hash((self.UserName, self.UserPath, super().__hash__()))


class Group(IdentityWithPolicies):
    def __init__(self, group_name, group_path):
        self.GroupName = group_name
        self.GroupPath = group_path
        super(Group, self).__init__()

    def __eq__(self, other):
        if not isinstance(other, Group):
            return False

        return self.GroupName == other.GroupName and\
            self.GroupPath == other.GroupPath and\
            super().__eq__(other)

    def __lt__(self, other):
        return self.GroupName < other.GroupName

    def __hash__(self):
        return hash((self.GroupName, self.GroupPath, super().__hash__()))


class Role(IdentityWithPolicies):
    def __init__(self, role_name, role_path, trust_policy):
        self.RoleName = role_name
        self.RolePath = role_path
        self.TrustPolicy = trust_policy
        super(Role, self).__init__()

    def __eq__(self, other):
        if not isinstance(other, Role):
            return False

        return self.RoleName == other.RoleName and\
            self.RolePath == other.RolePath and\
            super().__eq__(other)

    def __lt__(self, other):
        return self.RoleName < other.RoleName

    def __hash__(self):
        return hash((self.RoleName, self.RolePath, super().__hash__()))


class PermissionSet(IdentityWithPolicies):
    def __init__(self, name):
        self.Name = name
        super(PermissionSet, self).__init__()

    def __eq__(self, other):
        if not isinstance(other, PermissionSet):
            return False

        return self.Name == other.Name and\
            super().__eq__(other)

    def __lt__(self, other):
        return self.Name < other.Name

    def __hash__(self):
        return hash((self.Name, super().__hash__()))


class Policy:
    def __init__(self, name, document, path="/", is_aws_managed_policy=False):
        self.Name = name
        self.Policy = document
        self.Path = path
        self.IsAWSManagedPolicy = is_aws_managed_policy

    def __eq__(self, other):
        if not isinstance(other, Policy):
            return False

        names_are_equal = self.Name == other.Name
        paths_are_equal = self.Path == other.Path
        policies_are_equal = self.Policy == other.Policy
        is_managed_policy_is_equal = self.IsAWSManagedPolicy == other.IsAWSManagedPolicy

        return names_are_equal and paths_are_equal and policies_are_equal and is_managed_policy_is_equal

    def __lt__(self, other):
        return self.Name < other.Name

    def __hash__(self):
        return hash((self.Name, self.Path))


class Resource:
    def __init__(self, resource_name, resource_type, policy: Policy, configuration=None):
        if configuration is None:
            configuration = {}

        self.ResourceName = resource_name
        self.ResourceType = resource_type
        self.Policy = policy
        self.Configuration = configuration

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return False

        names_are_equal = self.ResourceName == other.ResourceName
        types_are_equal = self.ResourceType == other.ResourceType
        policies_are_equal = self.Policy == other.Policy
        configuration_is_equal = self.Configuration == other.Configuration

        return names_are_equal and types_are_equal and policies_are_equal and configuration_is_equal

    def custom_to_json(self):
        self_as_dict = self.__dict__
        if len(self_as_dict['Configuration']) == 0:
            self_as_dict.pop('Configuration')

        return self_as_dict

    def __lt__(self, other):
        return self.ResourceName < other.ResourceName

    def __hash__(self):
        return hash((self.ResourceName, self.ResourceType, self.Policy))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.cfn_tools.common_schema import string_schema
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema


class SplitEvaluator:
	def __init__(self, node_evaluator):
		self.node_evaluator = node_evaluator

	def evaluate(self, value, visited_values):
		validate_schema(value, split_schema, f'Fn::Split')

		delimiter = value[0]
		string_to_split = self.node_evaluator.eval_with_validation(value[1], string_schema, visited_values)

		return string_to_split.split(delimiter)


split_schema = {
	'type': 'array',
	'items': [
		{
			'type': 'string'
		},
		{}
	],
	'additionalItems': False
}


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.parsers.output import Policy, Resource


class SecretsManagerPolicyParser:
    """ AWS::SecretsManager::ResourcePolicy
    """
    def __init__(self):
        self.policies = []

    def parse(self, _, resource):
        evaluated_resource = resource.eval(secret_resource_policy_schema)
        properties = evaluated_resource['Properties']

        policy_document = properties['ResourcePolicy']
        secret_arn = properties['SecretId']

        try:
            secret_name = secret_arn.split("secret:", 1)[1]
        except IndexError:
            raise ApplicationError(f'Invalid value for {resource.ancestors_as_string()}.Properties.SecretId. Must be a valid Secret ARN. SecretId value: {secret_arn}')

        policy = Policy('ResourcePolicy', policy_document)
        resource = Resource(secret_name, 'AWS::SecretsManager::Secret', policy)

        self.policies.append(resource)

    def get_policies(self):
        return self.policies


secret_resource_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'ResourcePolicy': {
                    'type': 'object'
                },
                'SecretId': {
                    'type': 'string'
                }
            },
            'required': ['ResourcePolicy', 'SecretId']
        }
    },
    'required': ['Properties']
}


url_suffixes = {
	'cn-north-1': 'amazonaws.com.cn',
	'cn-northwest-1': 'amazonaws.com.cn'
}

default_url_suffix = 'amazonaws.com'


def evaluate(region):
	return url_suffixes.get(region, default_url_suffix)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator import client
from cfn_policy_validator.application_error import ApplicationError


class ImportValueEvaluator:
	def __init__(self, node_evaluator, region):
		self.node_evaluator = node_evaluator
		self.cloudformation_client = client.build('cloudformation', region)

	def evaluate(self, value, visited_values=None):
		if visited_values is None:
			visited_values = []

		imported_value = self.node_evaluator.eval(value, visited_values=visited_values)

		paginator = self.cloudformation_client.get_paginator('list_exports')
		response_iterator = paginator.paginate()
		for page in response_iterator:
			exports = page['Exports']
			for export in exports:
				if export['Name'] == imported_value:
					return export['Value']

		raise ApplicationError(f'Unable to resolve Fn::ImportValue. Could not find a stack export to import with value {imported_value}.')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.cfn_tools.common_schema import array_of_strings_schema
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema


class JoinEvaluator:
	def __init__(self, node_evaluator):
		self.node_evaluator = node_evaluator

	def evaluate(self, value, visited_values):
		validate_schema(value, join_schema, 'Fn::Join')

		delimiter = value[0]
		list_of_values = value[1]

		list_of_evaluated_values = self.node_evaluator.eval_with_validation(list_of_values, array_of_strings_schema,
																			path='Fn::Join.1', visited_values=visited_values)

		return delimiter.join(list_of_evaluated_values)


join_schema = {
	'type': 'array',
	'items': [
		{
			'type': 'string'
		},
		{}
	],
	'additionalItems': False
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
class ApplicationError(Exception):
    pass


class SchemaValidationError(ApplicationError):
    pass



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import jsonschema

from jsonschema import ValidationError
from cfn_policy_validator.application_error import SchemaValidationError

alphanumeric_regex = '^[a-zA-Z0-9]+$'


def validate(template):
	# this schema validates that the CloudFormation template has a valid Resources, Parameters, and Mappings section.
	# validating upfront allows us to make assumptions later that we're dealing with a valid template
	template_schema = {
		'type': 'object',
		'properties': {
			'Resources': {
				'type': 'object',
				'patternProperties': {
					alphanumeric_regex: {
						'type': 'object',
						'properties': {
							'Type': {
								'type': 'string'
							},
							'Properties': {
								'type': 'object'
							}
						},
						'required': ['Type']
					}
				},
				'additionalProperties': False
			},
			'Parameters': {
				'type': 'object',
				'additionalProperties': {
					'type': 'object'
				}
			},
			'Mappings': {
				'type': 'object',
				'patternProperties': {
					alphanumeric_regex: {
						'type': 'object',
						'patternProperties': {
							'^.*$': {
								'type': 'object',
								'patternProperties': {
									'^.*$': {
										'type': ['string', 'array'],
										'minProperties': 1
									}
								},
								'minProperties': 1,
								'additionalProperties': False
							}
						},
						'minProperties': 1,
						'additionalProperties': False
					}
				},
				'additionalProperties': False
			}
		},
		'required': ['Resources']
	}

	validate_schema(template, template_schema)


def validate_schema(instance, schema, parent_path=None):
	try:
		jsonschema.validate(instance=instance, schema=schema)
	except ValidationError as e:
		message = e.message

		# try to print a useful error message about where the validation issue is in the template
		path = ''
		if e.path:
			path = e.path.popleft()
			for item in e.path:
				# items can be integers or strings
				path = f'{path}.{item}'

		if parent_path is not None:
			path = parent_path if path == '' else f'{parent_path}.{path}'

		if path:
			raise SchemaValidationError(f'{message}, Path: {path}')

		raise SchemaValidationError(f'{message}')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import datetime
import json


def default_to_json(value):
	if isinstance(value, datetime.date):
		return value.isoformat()
	else:
		return value.__dict__


class InvalidPolicyException(Exception):
	def __init__(self, message, policy):
		self.message = message
		self.policy = policy

	def to_string(self):
		return f'{self.message}\n{json.dumps(self.policy, default=default_to_json, indent=4)}'



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import io
import logging

from cfn_policy_validator import client, parameters
from cfn_policy_validator.application_error import SchemaValidationError, ApplicationError
from cfn_policy_validator.argument_actions import parse_findings_to_ignore, parse_allow_external_principals
from cfn_policy_validator.cfn_tools import cfn_loader
from cfn_policy_validator.parameters import validate_finding_types
from cfn_policy_validator.parsers.account_config import AccountConfig
from cfn_policy_validator.parsers.identity import IdentityParser
from cfn_policy_validator.parsers.output import Output
from cfn_policy_validator.parsers.resource.parser import ResourceParser
from cfn_policy_validator.validation import validator
from cfn_policy_validator.validation.reporter import default_finding_types_that_are_blocking

LOGGER = logging.getLogger('cfn-policy-validator')


def validate(template_body,
             region,
             account_id,
             partition,
             template_parameters=None,
             ignore_finding=None,
             treat_as_blocking=default_finding_types_that_are_blocking,
             allowed_external_principals=None,
             allow_dynamic_ref_without_version=False):
    """
    Parses a CloudFormation template and runs it through IAM Access Analyzer for validation.
    @param template_body: String containing the body of the CloudFormation template.
    @param region: The region that the CloudFormation template will be deployed to.
    @param account_id: The AWS account ID that the CloudFormation template will be deployed to.
    @param partition: The AWS partition that the CloudFormation template will be deployed to.
    @param template_parameters:  A key: value dictionary of parameters that will be passed to the CloudFormation
        template when deployed.  e.g. { 'Parameter1Name': 'Parameter1Value', 'Parameter2Name': 'Parameter2Value' }
    @param ignore_finding: Allow validation failures to be ignored. Specify as a list of findings to be ignored.
        Can be individual finding codes (e.g. "PASS_ROLE_WITH_STAR_IN_RESOURCE"), a specific resource name
        (e.g. "MyResource"), or a combination of both separated by a period.
        (e.g. "MyResource.PASS_ROLE_WITH_STAR_IN_RESOURCE").
        Example: ["MyResource", "PASS_ROLE_WITH_STAR_IN_RESOURCE"]
    @param treat_as_blocking: Specify which finding types should be treated as errors. Other finding types are treated
        as warnings. Defaults to "ERROR" and "SECURITY_WARNING". Specify as a list of finding types that should be
        blocking. Possible values are "ERROR", "SECURITY_WARNING", "SUGGESTION", and "WARNING".  Pass "NONE" to ignore
        all errors.
        Example: ["ERROR", "SECURITY_WARNING", "WARNING"]
    @param allowed_external_principals: A list of external principals that should be ignored.  Specify as a list of
        a 12 digit AWS account ID, a federated web identity user, a federated SAML user, or an ARN. Specify "*"
        to allow anonymous access.
        Example: ["123456789123","arn:aws:iam::111111111111:role/MyOtherRole","graph.facebook.com"]
    @param allow_dynamic_ref_without_version: A boolean to allow the retrieval of dynamic references without specifying a version number. Specifying a version number
        helps ensure that the reference does not change between validation and deployment. Allowing dynamic references without versions would make it
        possible for the template you deploy to be different from the one that was validated.
    @return: A JSON formatted object containing findings classified as either blocking or non-blocking from IAM Access
        Analyzer
    """

    # this is done to avoid mutable default arguments as described here:
    # https://docs.python-guide.org/writing/gotchas/
    template_parameters = {} if template_parameters is None else template_parameters

    # validate input parameters
    ignore_finding = parse_findings_to_ignore(ignore_finding)
    allowed_external_principals = parse_allow_external_principals(allowed_external_principals)
    treat_as_blocking = validate_finding_types(treat_as_blocking)

    return _inner_validate(template_body, region, account_id, partition,
                           template_parameters, ignore_finding, treat_as_blocking, allowed_external_principals, allow_dynamic_ref_without_version)


def _inner_validate(template_body, region, account_id, partition,
                    template_parameters, ignore_finding, treat_as_blocking, allowed_external_principals,
                    allow_dynamic_ref_without_version: bool):
    account_config = AccountConfig(partition, region, account_id)
    template = _parse_template(template_body, account_config, template_parameters, allow_dynamic_ref_without_version)
    parser_output = _parse_template_output(template, account_config)
    report = validator.validate(parser_output, ignore_finding, treat_as_blocking, allowed_external_principals)

    return report.to_json()


def parse(template_body,
          region,
          account_id,
          partition,
          template_parameters=None,
          allow_dynamic_ref_without_version=False):
    """
        Parses a CloudFormation template.
        @param template_body: String containing the body of the CloudFormation template.
        @param region: The region that the CloudFormation template will be deployed to.
        @param account_id: The AWS account ID that the CloudFormation template will be deployed to.
        @param partition: The AWS partition that the CloudFormation template will be deployed to.
        @param template_parameters:  A key: value dictionary of parameters that will be passed to the CloudFormation
            template when deployed.  e.g. { 'Parameter1Name': 'Parameter1Value', 'Parameter2Name': 'Parameter2Value' }
        @param allow_dynamic_ref_without_version: A boolean to allow the retrieval of dynamic references without specifying a version number. Specifying a version number
            helps ensure that the reference does not change between validation and deployment. Allowing dynamic references without versions would make it
            possible for the template you deploy to be different from the one that was validated.
        @return: A JSON formatted object containing findings classified as either blocking or non-blocking from IAM Access
            Analyzer
        """
    template_parameters = {} if template_parameters is None else template_parameters

    # TODO: deal with logging
    return _inner_parse(template_body, region, account_id, partition, template_parameters, allow_dynamic_ref_without_version)


def _inner_parse(template_body, region, account_id, partition, template_parameters, allow_dynamic_ref_without_version: bool):
    account_config = AccountConfig(partition, region, account_id)
    template = _parse_template(template_body, account_config, template_parameters, allow_dynamic_ref_without_version)
    parser_output = _parse_template_output(template, account_config)

    return parser_output.to_json()


def _parse_template(template_body, account_config, template_parameters, allow_dynamic_ref_without_version: bool):
    stream = io.StringIO(template_body)

    try:
        template = cfn_loader.load(stream, account_config, template_parameters, allow_dynamic_ref_without_version)
    except SchemaValidationError:
        logging.exception('Unable to parse CloudFormation template.  Invalid CloudFormation schema detected.')
        raise ApplicationError('Unable to parse CloudFormation template.  Invalid CloudFormation schema detected.')
    except Exception:
        logging.exception('Unable to parse CloudFormation template.  Invalid YAML or JSON detected.')
        raise ApplicationError('Unable to parse CloudFormation template.  Invalid YAML or JSON detected.')

    return template


def _parse_template_file(file_path, account_config, template_parameters, allow_dynamic_ref_without_version: bool):
    try:
        with open(file_path, 'r') as stream:
            try:
                template = cfn_loader.load(stream, account_config, template_parameters, allow_dynamic_ref_without_version)
            except SchemaValidationError:
                logging.exception('Unable to parse CloudFormation template.  Invalid CloudFormation schema detected.')
                raise ApplicationError('Unable to parse CloudFormation template.  Invalid CloudFormation schema detected.')
            except Exception:
                logging.exception('Unable to parse CloudFormation template.  Invalid YAML or JSON detected.')
                raise ApplicationError('Unable to parse CloudFormation template.  Invalid YAML or JSON detected.')
    except FileNotFoundError:
        raise ApplicationError(f'CloudFormation template not found: {file_path}')

    return template


def _parse_template_output(template, account_config):
    output = Output(account_config)
    output.Roles, output.Users, output.Groups, output.PermissionSets, output.OrphanedPolicies = \
        IdentityParser.parse(template, account_config)
    output.Resources = ResourceParser.parse(template, account_config)

    return output



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import yaml

from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.cfn_object import CfnObject


TAG_MAP = "tag:yaml.org,2002:map"
TAG_TIMESTAMP = "tag:yaml.org,2002:timestamp"
UNCONVERTED_SUFFIXES = ["Ref", "Condition"]
FN_PREFIX = "Fn::"


class CfnYamlLoader(yaml.SafeLoader):
    def multi_constructor(self, tag_suffix, node):
        if tag_suffix not in UNCONVERTED_SUFFIXES:
            tag_suffix = "{}{}".format(FN_PREFIX, tag_suffix)

        if tag_suffix == "Fn::GetAtt":
            constructor = self.construct_getatt
        elif isinstance(node, yaml.ScalarNode):
            constructor = self.construct_scalar
        elif isinstance(node, yaml.SequenceNode):
            constructor = self.construct_sequence
        elif isinstance(node, yaml.MappingNode):
            constructor = self.construct_mapping
        else:
            raise Exception("Bad tag: !{}".format(tag_suffix))

        return CfnObject((
            (tag_suffix, constructor(node)),
        ))

    def construct_getatt(self, node):
        if isinstance(node.value, str):
            return node.value.split(".", 1)
        if isinstance(node.value, list):
            return [s.value for s in node.value]
        else:
            raise ApplicationError("Unexpected node type: {}".format(type(node.value)))

    def construct_mapping(self, node, deep=False):
        mapping = CfnObject()

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            value = self.construct_object(value_node, deep=deep)

            mapping[key] = value

        return mapping

    # for consistency with JSON parsing, parse datetime to string
    # must include self as first argument
    # noinspection PyMethodMayBeStatic
    def parse_datetime(self, node):
        return node.value


CfnYamlLoader.add_constructor(TAG_MAP, CfnYamlLoader.construct_mapping)
CfnYamlLoader.add_constructor(TAG_TIMESTAMP, CfnYamlLoader.parse_datetime)
CfnYamlLoader.add_multi_constructor("!", CfnYamlLoader.multi_constructor)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
import logging
import re
import time
import uuid

from cfn_policy_validator.canonical_user_id import client, get_canonical_user
from cfn_policy_validator.validation.findings import Findings
from cfn_policy_validator.validation import default_to_json, InvalidPolicyException
from cfn_policy_validator.validation.reporter import Reporter
from cfn_policy_validator.application_error import ApplicationError


LOGGER = logging.getLogger("cfn-policy-validator")


def validate(parser_output, findings_to_ignore, finding_types_that_are_errors, allowed_external_principals):
	"""
	Run the output from the parsers through IAM Access Analyzer, filter, and report the findings.
	"""

	findings = validate_parser_output(parser_output)

	reporter = Reporter(findings_to_ignore, finding_types_that_are_errors, allowed_external_principals)
	return reporter.build_report_from(findings)


def validate_parser_output(parser_output):
	"""
	Run the parser output through IAM Access Analyzer validation
	"""
	validator = Validator(parser_output.Account, parser_output.Region, parser_output.Partition)

	validator.validate_policies(parser_output.OrphanedPolicies)
	validator.validate_roles(parser_output.Roles)
	validator.validate_users(parser_output.Users)
	validator.validate_groups(parser_output.Groups)
	validator.validate_permission_sets(parser_output.PermissionSets)
	validator.validate_resources(parser_output.Resources)

	return validator.findings


class Validator:
	RESOURCE_POLICY_TYPE = 'RESOURCE_POLICY'
	IDENTITY_POLICY_TYPE = 'IDENTITY_POLICY'

	def __init__(self, account_id, region, partition):
		self.findings = Findings()
		self.access_analyzer_name = 'AnalyzerCreatedByCfnIAMPolicyValidator'
		self.analyzer_arn = None

		self.client = client.build('accessanalyzer', region)

		# preview builders are used to build the access preview configuration for an individual resource type
		# a preview builder must be added to add support for access previews for a given resource
		self.preview_builders = {
			'AWS::SQS::Queue': SqsQueuePreviewBuilder(account_id, region, partition),
			'AWS::KMS::Key': KmsKeyPreviewBuilder(account_id, region, partition),
			'AWS::S3::AccessPoint': S3SingleRegionAccessPointPreviewBuilder(account_id, region, partition),
			'AWS::S3::MultiRegionAccessPoint': S3MultiRegionAccessPointPreviewBuilder(account_id, partition),
			'AWS::S3::Bucket': S3BucketPreviewBuilder(region, partition),
			'AWS::IAM::Role::TrustPolicy': RoleTrustPolicyPreviewBuilder(account_id, partition),
			'AWS::SecretsManager::Secret': SecretsManagerSecretPreviewBuilder(account_id, region, partition)
		}

		# maps the resource type to the parameter for validate_policy that enables service specific policy validation
		# not all services have service specific policy validation.  The names may be identical for now, but we don't
		# want to rely on that
		self.service_specific_policy_validation = {
			'AWS::S3::Bucket': 'AWS::S3::Bucket',
			'AWS::S3::AccessPoint': 'AWS::S3::AccessPoint',
			'AWS::S3::MultiRegionAccessPoint': 'AWS::S3::MultiRegionAccessPoint'
		}

		self.maximum_number_of_access_preview_attempts = 150
		self._try_create_analyzer()

	def validate_roles(self, roles):
		"""
		Validate policies attached to roles
		"""
		previews_to_await = []
		for role in roles:
			LOGGER.info(f'Validating trust policy for role {role.RoleName}..')
			response = self._validate_policy(role.TrustPolicy, self.RESOURCE_POLICY_TYPE, False)
			LOGGER.info(f'ValidatePolicy response: {response}')

			validation_findings = response['findings']
			self.findings.add_validation_finding(validation_findings, role.RoleName, 'TrustPolicy')

			# use access previews to validate a role's trust policy
			preview = self.__validate_role_trust_policy(role, validation_findings)
			previews_to_await.append(preview)

			# validate identity policies attached to the role
			for policy in role.Policies:
				LOGGER.info(f'Validating identity policy for {role.RoleName} with name {policy.Name}')
				response = self._validate_policy(policy.Policy, self.IDENTITY_POLICY_TYPE, policy.IsAWSManagedPolicy)
				LOGGER.info(f'ValidatePolicy response: {response}')
				self.findings.add_validation_finding(response['findings'], role.RoleName, policy.Name)

		access_preview_findings = self._wait_for_findings(previews_to_await)
		for access_preview_finding in access_preview_findings:
			self.findings.add_trust_policy_finding(access_preview_finding.findings, access_preview_finding.resource.RoleName)

	def __validate_role_trust_policy(self, role, validation_findings):
		preview_builder = self.preview_builders['AWS::IAM::Role::TrustPolicy']
		configuration = preview_builder.build_configuration(role)

		LOGGER.info(f'Creating access preview with configuration {configuration}')
		response = self.client.create_access_preview(
			analyzerArn=self.analyzer_arn,
			configurations=configuration
		)
		LOGGER.info(f'CreateAccessPreview response: {response}')
		return PreviewAwaitingResponse(response['id'], role, role.RoleName, validation_findings)

	def validate_policies(self, policies):
		"""
		Validate orphaned policies
		"""
		resource_name = 'No resource attached'
		for policy in policies:
			LOGGER.info(f'Validating identity policy for {policy.Name}')
			response = self._validate_policy(policy.Policy, self.IDENTITY_POLICY_TYPE, policy.IsAWSManagedPolicy)
			LOGGER.info(f'ValidatePolicy response: {response}')
			self.findings.add_validation_finding(response['findings'], resource_name, policy.Name)

	def validate_users(self, users):
		"""
		Validate policies attached to users
		"""
		for user in users:
			for policy in user.Policies:
				LOGGER.info(f'Validating identity policy for user {user.UserName} with policy name {policy.Name}')
				response = self._validate_policy(policy.Policy, self.IDENTITY_POLICY_TYPE, policy.IsAWSManagedPolicy)
				LOGGER.info(f'ValidatePolicy response {response}')
				self.findings.add_validation_finding(response['findings'], user.UserName, policy.Name)

	def validate_groups(self, groups):
		"""
		Validate policies attached to groups
		"""
		for group in groups:
			for policy in group.Policies:
				LOGGER.info(f'Validating identity policy for group {group.GroupName} with policy name {policy.Name}')
				response = self._validate_policy(policy.Policy, self.IDENTITY_POLICY_TYPE, policy.IsAWSManagedPolicy)
				LOGGER.info(f'ValidatePolicy response {response}')
				self.findings.add_validation_finding(response['findings'], group.GroupName, policy.Name)

	def validate_permission_sets(self, permission_sets):
		"""
		Validate policies attached to permission sets
		"""
		for permission_set in permission_sets:
			for policy in permission_set.Policies:
				LOGGER.info(f'Validating identity policy for permission set {permission_set.Name} with policy name {policy.Name}')
				response = self._validate_policy(policy.Policy, self.IDENTITY_POLICY_TYPE, policy.IsAWSManagedPolicy)
				LOGGER.info(f'ValidatePolicy response {response}')
				self.findings.add_validation_finding(response['findings'], permission_set.Name, policy.Name)

	def validate_resources(self, resources):
		"""
		Validate resource policies
		"""
		previews_to_await = []
		for resource in resources:
			validation_findings = []
			if resource.Policy.Policy is None:
				LOGGER.info(f'Resource {resource.ResourceName} has no resource-based policy.  Skipping call to ValidatePolicy.')
			else:
				# we want to run validate_policy on all resource policies regardless of if they are supported policies
				# for access previews
				LOGGER.info(f'Validating resource policy for resource {resource.ResourceName} of type {resource.ResourceType}')

				validate_policy_resource_type = self.service_specific_policy_validation.get(resource.ResourceType)
				response = self._validate_policy(resource.Policy.Policy, self.RESOURCE_POLICY_TYPE, False, validate_policy_resource_type)
				LOGGER.info(f'ValidatePolicy response {response}')
				validation_findings = response['findings']
				self.findings.add_validation_finding(validation_findings, resource.ResourceName, resource.Policy.Name)

			# only supported policies for access previews will have config builders
			preview_builder = self.preview_builders.get(resource.ResourceType)
			if preview_builder is not None:
				try:
					configuration = preview_builder.build_configuration(resource)
				except InvalidPolicyException as e:
					self._raise_invalid_configuration_error_for(resource.ResourceName, validation_findings, e.to_string())

				LOGGER.info(f'Creating access preview for resource {resource.ResourceName} of type {resource.ResourceType}')
				LOGGER.info(f'Using access preview configuration: {configuration}')

				try:
					response = self.client.create_access_preview(
						analyzerArn=self.analyzer_arn,
						configurations=configuration
					)
				except Exception as e:
					raise ApplicationError(f'Failed to create access preview for {resource.ResourceName}.', e)

				LOGGER.info(f'CreateAccessPreview response: {response}')
				preview = PreviewAwaitingResponse(response['id'], resource, resource.ResourceName, validation_findings)
				previews_to_await.append(preview)

		# batch and wait for all access previews to complete
		access_preview_findings = self._wait_for_findings(previews_to_await)

		for access_preview_finding in access_preview_findings:
			self.findings.add_external_principal_finding(access_preview_finding.findings,
														 access_preview_finding.resource.ResourceName,
														 access_preview_finding.resource.Policy.Name)

	def _validate_policy(self, policy_as_json, policy_type, is_aws_managed_policy, resource_type=None):
		policy_as_string = json.dumps(policy_as_json)
		policy_without_whitespace = policy_as_string.replace(" ", "")
		length_of_policy_without_whitespace = len(policy_without_whitespace)
		if length_of_policy_without_whitespace >= 32768:
			if is_aws_managed_policy:
				LOGGER.info(f'AWS managed policy has length {length_of_policy_without_whitespace} which exceeds the max size (32768 bytes) for ValidatePolicy call.  Ignoring..')
				return {'findings': []}
			else:
				# if this is a customer policy, we return an error finding
				return {
					'findings': [self._build_policy_size_finding(length_of_policy_without_whitespace)]
				}

		if resource_type is None:
			response = self.client.validate_policy(
				policyType=policy_type,
				policyDocument=policy_as_string
			)
		else:
			LOGGER.info(f'Running service specific policy validation for {resource_type}')
			response = self.client.validate_policy(
				policyType=self.RESOURCE_POLICY_TYPE,
				policyDocument=policy_as_string,
				validatePolicyResourceType=resource_type
			)

		return response

	@staticmethod
	def _build_policy_size_finding(policy_size):
		return {
			'findingType': 'ERROR',
			'findingDetails': f'The {policy_size} characters in the policy, excluding whitespace, exceed the character maximum for the ValidatePolicy API. We recommend that you use multiple granular policies.',
			'issueCode': 'POLICY_SIZE_EXCEEDS_VALIDATE_POLICY_MAXIMUM'
		}

	def _wait_for_findings(self, previews_to_await):
		findings = []
		for preview in previews_to_await:
			number_of_attempts = 0
			while True:
				LOGGER.info(f'Waiting on access preview {preview.id} to finish creating attempt {number_of_attempts+1}..')
				response = self.client.get_access_preview(
					accessPreviewId=preview.id,
					analyzerArn=self.analyzer_arn
				)
				LOGGER.info(f'GetAccessPreview response: {response}')
				status = response['accessPreview']['status']

				if status == 'CREATING':
					number_of_attempts = number_of_attempts + 1
					if number_of_attempts >= self.maximum_number_of_access_preview_attempts:
						raise ApplicationError(f'Timed out after 5 minutes waiting for access analyzer preview to create.')

					time.sleep(2)
				else:
					break

			LOGGER.info(f'Access preview creation completed for {preview.name} with status {status}')

			if status == 'FAILED':
				reason = response["accessPreview"]["statusReason"]["code"]
				if reason == 'INVALID_CONFIGURATION':
					self._raise_invalid_configuration_error_for(preview.name, preview.validation_findings)

				raise ApplicationError(f'Failed to create access preview for {preview.name}.  Reason: {reason}')

			paginator = self.client.get_paginator('list_access_preview_findings')
			for page in paginator.paginate(accessPreviewId=preview.id, analyzerArn=self.analyzer_arn):
				active_findings = []
				for finding in page['findings']:
					LOGGER.info(f'Access preview finding: {finding}')
					if finding['status'] == 'ACTIVE':
						active_findings.append(finding)

				if len(active_findings) > 0:
					findings.append(AccessPreviewFindings(preview.resource, active_findings))

		return findings

	@staticmethod
	def _filter_findings_for_source_type(raw_findings, desired_source_type):
		filtered_findings = []
		for raw_finding in raw_findings:
			sources = raw_finding.get('sources')
			if sources is None:
				continue

			if any([source for source in sources if source.get('type') == desired_source_type]):
				filtered_findings.append(raw_finding)

		return filtered_findings

	def _try_create_analyzer(self):
		if self.analyzer_arn is not None:
			return

		response = self.client.list_analyzers(
			type='ACCOUNT'
		)

		first_active_analyzer = next((analyzer for analyzer in response['analyzers'] if analyzer['status'] == 'ACTIVE'), None)
		if first_active_analyzer is not None:
			self.analyzer_arn = first_active_analyzer['arn']
			return

		LOGGER.info('No active analyzers found in account.  Creating analyzer.')
		response = self.client.create_analyzer(
			analyzerName=self.access_analyzer_name,
			type='ACCOUNT'
		)
		self.analyzer_arn = response['arn']

	@staticmethod
	def _raise_invalid_configuration_error_for(resource_name, validation_findings, custom_message=None):
		# if we get an invalid configuration error, surface the validation findings as they likely point
		# to the issue
		message = f'Failed to create access preview for {resource_name}.  Validate that your trust or resource policy\'s ' \
				  f'schema is correct.'
		if len(validation_findings) > 0:
			message += f'\nThe following validation findings were detected for this resource: ' \
					   f'{json.dumps(validation_findings, default=default_to_json, indent=4)}.'

		if custom_message is not None:
			message += f'\nAdditional information:\n{custom_message}'

		raise ApplicationError(message)


class AccessPreviewFindings:
	def __init__(self, resource, findings):
		self.resource = resource
		self.findings = findings


class PreviewAwaitingResponse:
	def __init__(self, preview_id, resource, name, validation_findings):
		self.id = preview_id
		self.resource = resource
		self.name = name
		self.validation_findings = validation_findings


class SqsQueuePreviewBuilder:
	def __init__(self, account_id, region, partition):
		self.region = region
		self.account_id = account_id
		self.partition = partition

	def build_configuration(self, resource):
		policy = json.dumps(resource.Policy.Policy)

		return {
			f'arn:{self.partition}:sqs:{self.region}:{self.account_id}:{resource.ResourceName}': {
				'sqsQueue': {
					'queuePolicy': policy
				}
			}
		}


class KmsKeyPreviewBuilder:
	def __init__(self, account_id, region, partition):
		self.account_id = account_id
		self.region = region
		self.partition = partition

	def build_configuration(self, resource):
		policy = json.dumps(resource.Policy.Policy)

		return {
			f'arn:{self.partition}:kms:{self.region}:{self.account_id}:key/{resource.ResourceName}': {
				'kmsKey': {
					'keyPolicies': {
						'default': policy
					}
				}
			}
		}


class S3BucketPreviewBuilder:
	def __init__(self, region, partition):
		self.partition = partition
		self.region = region

	def build_configuration(self, resource):
		s3_bucket_json = {}

		# either policy or access control must be present, there exists no resource where they are both null
		if resource.Policy.Policy is not None:
			policy = json.dumps(resource.Policy.Policy)
			s3_bucket_json['bucketPolicy'] = policy

		access_control = resource.Configuration.get('AccessControl')
		if access_control is not None:
			bucket_acl_grants = self.__build_bucket_acl_grants(resource.ResourceName, access_control)
			s3_bucket_json['bucketAclGrants'] = bucket_acl_grants

		return {
			f'arn:{self.partition}:s3:::{resource.ResourceName}': {
				's3Bucket': s3_bucket_json
			}
		}

	def __build_bucket_acl_grants(self, bucket_name, access_control_value):
		if access_control_value == 'Private':
			return [self.__build_owner_full_control_grant()]
		elif access_control_value == 'PublicRead':
			return [
				self.__build_owner_full_control_grant(),
				self.__build_grant_uri(self.all_users_group_uri, 'READ')
			]
		elif access_control_value == 'PublicReadWrite':
			return [
				self.__build_owner_full_control_grant(),
				self.__build_grant_uri(self.all_users_group_uri, 'READ'),
				self.__build_grant_uri(self.all_users_group_uri, 'WRITE')
			]
		elif access_control_value == 'AuthenticatedRead':
			return [
				self.__build_owner_full_control_grant(),
				self.__build_grant_uri(self.authenticated_users_group_uri, 'READ')
			]
		elif access_control_value == 'LogDeliveryWrite':
			return [
				self.__build_grant_uri(self.log_delivery_group_uri, 'WRITE'),
				self.__build_grant_uri(self.log_delivery_group_uri, 'READ_ACP')
			]
		elif access_control_value == 'BucketOwnerRead':
			return [self.__build_owner_full_control_grant()]
		elif access_control_value == 'BucketOwnerFullControl':
			return [self.__build_owner_full_control_grant()]
		elif access_control_value == 'AwsExecRead':
			return [
				self.__build_owner_full_control_grant(),
				self.__build_grant_id(self.ec2_canonical_id, 'READ')
			]
		else:
			raise ApplicationError(f'Invalid AccessControl value "{access_control_value}" for {bucket_name}.\n'
								f'See https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-s3-bucket.html#cfn-s3-bucket-accesscontrol '
								f'for valid AccessControl values.')

	all_users_group_uri = 'http://acs.amazonaws.com/groups/global/AllUsers'
	authenticated_users_group_uri = 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers'
	log_delivery_group_uri = 'http://acs.amazonaws.com/groups/s3/LogDelivery'
	ec2_canonical_id = '6aa5a366c34c1cbe25dc49211496e913e0351eb0e8c37aa3477e40942ec6b97c'

	@staticmethod
	def __build_grant_uri(uri, permission):
		return {
			'grantee': {
				'uri': uri
			},
			'permission': permission
		}

	@staticmethod
	def __build_grant_id(grant_id, permission):
		return {
			'grantee': {
				'id': grant_id
			},
			'permission': permission
		}

	def __build_owner_full_control_grant(self):
		return self.__build_grant_id(get_canonical_user(self.region), 'FULL_CONTROL')


class S3AccessPointPreviewBuilder:
	def __init__(self, account_id, partition, region=None):
		self.partition = partition
		self.account_id = account_id
		self.region = region

	# this bucket policy enforces that access must come through an access point so that we can evaluate the access point
	# policy independent of the bucket policy
	def build_bucket_policy(self, bucket_name):
		return {
			"Version": "2012-10-17",
			"Statement": [
				{
					"Effect": "Allow",
					"Principal": {
						"AWS": "*"
					},
					"Action": "*",
					"Resource": [
						f"arn:aws:s3:::{bucket_name}",
						f"arn:aws:s3:::{bucket_name}/*"
					],
					"Condition": {
						"StringEquals": {
							"s3:DataAccessPointAccount": self.account_id
						}
					}
				}
			]
		}


class S3MultiRegionAccessPointPreviewBuilder(S3AccessPointPreviewBuilder):
	def __init__(self, account_id, partition):
		super(S3MultiRegionAccessPointPreviewBuilder, self).__init__(account_id, partition)

	def build_configuration(self, resource):
		policy_json = resource.Policy.Policy
		policy = json.dumps(resource.Policy.Policy)

		# since we're evaluating the access point independently, the name of the bucket does not matter
		bucket_name = str(uuid.uuid4())

		bucket_policy_json = self.build_bucket_policy(bucket_name)
		bucket_policy = json.dumps(bucket_policy_json)

		access_point_arn = self.build_access_point_arn(resource.ResourceName, policy_json)

		return {
			f'arn:{self.partition}:s3:::{bucket_name}': {
				's3Bucket': {
					'accessPoints': {
						access_point_arn: {
							'accessPointPolicy': policy
						}
					},
					'bucketPolicy': bucket_policy
				}
			}
		}

	# match the alias of the multi region access point in an access point's ARN:
	# "arn:aws:s3::111111111111:accesspoint/MyAccessPoint.mrap/object/abc/*" would find "MyAccessPoint.mrap"
	access_point_alias_regex = re.compile(r'arn:[^:]*:s3::[^:]*:accesspoint/([^/]*)')

	# find the access point ARN from the first resource in the access point policy
	def build_access_point_arn(self, access_point_name, access_point_policy):
		error_message_prefix = f"Access point policy for {access_point_name}"
		statements = access_point_policy.get('Statement')
		if not isinstance(statements, list):
			statements = [statements]

		# we only need to look at the first resource in the first statement.  All statements must have resources
		# and all resources must include the multi region access point ARN.
		first_statement = next(iter(statements), None)
		if not isinstance(first_statement, dict):
			raise InvalidPolicyException(f"{error_message_prefix} has 'Statement' that is missing or is of invalid type.", access_point_policy)

		resource = first_statement.get('Resource')
		if not isinstance(resource, list):
			resource = [resource]

		resource = next(iter(resource), None)
		if not isinstance(resource, str):
			raise InvalidPolicyException(f"{error_message_prefix} has 'Resource' element that is missing or is of invalid type.", access_point_policy)

		match = self.access_point_alias_regex.search(resource)
		if match is None:
			raise InvalidPolicyException(f"{error_message_prefix} has entry for 'Resource' with invalid multi-region access point ARN.", access_point_policy)
		else:
			return f'arn:{self.partition}:s3::{self.account_id}:accesspoint/{match.group(1)}'


class S3SingleRegionAccessPointPreviewBuilder(S3AccessPointPreviewBuilder):
	def __init__(self, account_id, region, partition):
		super(S3SingleRegionAccessPointPreviewBuilder, self).__init__(account_id, partition, region)

	def build_configuration(self, resource):
		policy = json.dumps(resource.Policy.Policy)

		# since we're evaluating the access point independently, the name of the bucket does not matter
		bucket_name = str(uuid.uuid4())

		bucket_policy_json = self.build_bucket_policy(bucket_name)
		bucket_policy = json.dumps(bucket_policy_json)

		network_origin = {
			'internetConfiguration': {}
		}

		if resource.Configuration is not None and 'VpcId' in resource.Configuration:
			network_origin = {
				'vpcConfiguration': {
					'vpcId': resource.Configuration['VpcId']
				}
			}

		return {
			f'arn:{self.partition}:s3:::{bucket_name}': {
				's3Bucket': {
					'accessPoints': {
						f'arn:{self.partition}:s3:{self.region}:{self.account_id}:accesspoint/{resource.ResourceName}': {
							'accessPointPolicy': policy,
							'networkOrigin': network_origin
						}
					},
					'bucketPolicy': bucket_policy
				}
			}
		}


class RoleTrustPolicyPreviewBuilder:
	def __init__(self, account_id, partition):
		self.account_id = account_id
		self.partition = partition

	def build_configuration(self, resource):
		policy = json.dumps(resource.TrustPolicy)

		role_name = resource.RoleName[:64]

		return {
			f'arn:{self.partition}:iam::{self.account_id}:role/{role_name}': {
				'iamRole': {
					'trustPolicy': policy
				}
			}
		}


class SecretsManagerSecretPreviewBuilder:
	def __init__(self, account_id, region, partition):
		self.account_id = account_id
		self.region = region
		self.partition = partition

	def build_configuration(self, resource):
		policy = json.dumps(resource.Policy.Policy)

		return {
			# secrets manager arns have a random 6 characters appended to the end
			f'arn:{self.partition}:secretsmanager:{self.region}:{self.account_id}:secret:{resource.ResourceName}-3xyxqI': {
				'secretsManagerSecret': {
					'secretPolicy': policy
				}
			}
		}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
import os

from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema
from cfn_policy_validator.parsers.utils.cycle_detection import validate_no_cycle
from cfn_policy_validator.parsers.utils.intrinsic_functions import name_hints
from cfn_policy_validator.parsers.utils.intrinsic_functions import aws_url_suffix_evaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.aws_no_value_evaluator import NoValue


class RefEvaluator:
	def __init__(self, resources, arn_generator, parameters, parameter_values, account_config, node_evaluator):
		this_files_directory = os.path.dirname(os.path.realpath(__file__))
		with open(os.path.join(this_files_directory, '..', 'services_where_ref_returns_arn.json')) as f:
			services_where_ref_returns_arn = json.load(f)

		self.services_where_ref_returns_arn = services_where_ref_returns_arn['Resources']
		self.resources = resources
		self.arn_generator = arn_generator
		self.parameters = parameters
		self.parameter_values = parameter_values
		self.account_config = account_config
		self.node_evaluator = node_evaluator

		# some resources require custom evaluation logic, but we should only need to care about this for resources that
		# have resource policies that we want to parse
		self.custom_ref_evals = {
			'AWS::SQS::Queue': evaluate_sqs_queue_ref
		}

	def evaluate(self, resource_logical_name_or_param, visited_values=None):
		""" Evaluates a Fn::Ref function
			visited_values: tracks visited values to detect circular references in a CloudFormation template
		"""
		if visited_values is None:
			visited_values = []

		validate_schema(resource_logical_name_or_param, ref_schema, 'Ref')

		if resource_logical_name_or_param == "AWS::AccountId":
			return self.account_config.account_id

		if resource_logical_name_or_param == "AWS::Partition":
			return self.account_config.partition

		if resource_logical_name_or_param == "AWS::Region":
			return self.account_config.region

		if resource_logical_name_or_param == "AWS::StackName":
			# just return some default value, we won't know this in advance
			return "StackName"

		if resource_logical_name_or_param == "AWS::NoValue":
			return NoValue()

		if resource_logical_name_or_param == "AWS::URLSuffix":
			return aws_url_suffix_evaluator.evaluate(self.account_config.region)

		# check to see if the reference is for a resource
		resource = self.resources.get(resource_logical_name_or_param)
		if resource is not None:
			resource_type = resource['Type']

			# first, see if this ref should return an ARN.  Not all Ref's return ARNs.
			if resource_type in self.services_where_ref_returns_arn:
				return self.arn_generator.try_generate_arn(resource_logical_name_or_param, resource, 'Ref', visited_values=visited_values)

			# next, see if we have a custom evaluation for this ref
			custom_ref_eval = self.custom_ref_evals.get(resource_type)
			if custom_ref_eval is not None:
				return custom_ref_eval(resource_logical_name_or_param, resource, self.account_config, visited_values)

			# at this point, we make the assumption that the resource just returns a string (probably either the resource
			# name or ID)
			properties = resource.get('Properties', {})

			# next, attempt to return the actual name of the resource if one is defined and it's for a common resource
			name_returned_by_ref = name_hints.get(resource_type)
			property_value = properties.get(name_returned_by_ref)
			if property_value is None:
				# just return the CFN logical name if there's no property to be found
				return resource_logical_name_or_param

			# we found a valid property value for the name of the resources. this property may reference another resource,
			# so check to see if we've already done that and we're stuck in a cycle
			validate_no_cycle(resource_logical_name_or_param, name_returned_by_ref, visited_values)
			return self.node_evaluator.eval(property_value, visited_values=visited_values)

		# if it's not a reference to a resource, check to see if it references a parameter
		parameter = self.parameters.get(resource_logical_name_or_param)
		if parameter is not None:
			# if the parameter exists in the template, make sure a value was passed in for it
			parameter_value = self.parameter_values.get(resource_logical_name_or_param)
			if parameter_value is None:
				raise ApplicationError(f'No value passed for referenced parameter: {resource_logical_name_or_param}.\n'
									   f'Parameters are passed using the --parameters flag.')

			return parameter_value

		raise ApplicationError(f'Unable to find a referenced resource or parameter in template: {resource_logical_name_or_param}')


# used to map common CFN resource types to their name properties to more accurately generate names for common resources
# instead of defaulting to the CFN resource's logical name
# this is a user experience improvement and does not impact the validation / analysis of policies
ref_name_hints = {
	'AWS::S3::Bucket': 'BucketName',
	'AWS::Lambda::Function': 'FunctionName',
	'AWS::IAM::Role': 'RoleName',
	'AWS::IAM::User': 'UserName',
	'AWS::IAM::Group': 'GroupName',
	'AWS::SQS::Queue': 'QueueName'
}


ref_schema = {
	'type': 'string'
}


# for SQS, we need to know that Ref returns a queue URL.  This queue URL is used to link the queue's policy to the queue
def evaluate_sqs_queue_ref(resource_name, sqs_queue_resource, account_config, visited_values):
	evaluated_resource = sqs_queue_resource.eval(sqs_queue_schema, visited_values)

	properties = evaluated_resource.get('Properties', {})
	queue_name = properties.get('QueueName', resource_name)

	return f'https://sqs.{account_config.region}.amazonaws.com/{account_config.account_id}/{queue_name}'


sqs_queue_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'QueueName': {
					'type': 'string'
				}
			}
		}
	}
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import argparse
import logging
import sys
import traceback

from cfn_policy_validator.validation import validator
from cfn_policy_validator.version import __version__
from cfn_policy_validator import parameters, client, AccountConfig, \
    _parse_template_file, _parse_template_output
from cfn_policy_validator.validation.reporter import default_finding_types_that_are_blocking
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.argument_actions import DictionaryArgument, \
    ParseFindingsToIgnoreFromCLI, ParseAllowExternalPrincipalsFromCLI
from cfn_policy_validator.logger import configure_logging
from cfn_policy_validator.parameters import validate_region, validate_finding_types_from_cli, validate_credentials

LOGGER = logging.getLogger('cfn-policy-validator')


# consumable when running as CLI
def validate_from_cli(arguments):
    LOGGER.info(f'Validating template {arguments.template_path}')
    account_id, partition = client.get_account_and_partition(arguments.region)
    account_config = AccountConfig(partition, arguments.region, account_id)

    template = _parse_template_file(arguments.template_path, account_config, arguments.parameters, arguments.allow_dynamic_ref_without_version)
    parser_output = _parse_template_output(template, account_config)
    report = validator.validate(parser_output, arguments.ignore_finding, arguments.treat_as_blocking, arguments.allowed_external_principals)

    report.print()
    if report.has_blocking_findings():
        exit(2)
    else:
        exit(0)


# consumable when running as CLI
def parse_from_cli(arguments):
    LOGGER.info(f'Parsing template {arguments.template_path}')
    account_id, partition = client.get_account_and_partition(arguments.region)
    account_config = AccountConfig(partition, arguments.region, account_id)

    template = _parse_template_file(arguments.template_path, account_config, arguments.parameters, arguments.allow_dynamic_ref_without_version)
    parser_output = _parse_template_output(template, account_config)

    parser_output.print()
    exit(0)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--template-path', metavar="TEMPLATE_PATH", dest="template_path", required=True,
                        help='The path to the CloudFormation template.')

    parent_parser.add_argument('--region', dest="region", required=True, type=validate_region,
                        help="The region the resources will be deployed to.")

    parent_parser.add_argument('--parameters', action=DictionaryArgument, nargs="+", metavar="KEY=VALUE", dest="parameters",
                        help='Parameter key and value in the format -p Key1=Value1 Key2=Value2.  Only parameters'
                             ' that are referenced by IAM policies are required.', default={})

    parent_parser.add_argument('--template-configuration-file', metavar='FILE_PATH.json', dest="template_configuration_file",
                               help="A JSON formatted file that specifies template parameter values, a stack policy, and tags."
                                    "Everything but parameters are ignored from this file. Identical values passed in "
                                    "the --parameters flag override parameters in this file\n"
                                    "See CloudFormation documentation on format for this file: "
                                    "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/continuous-delivery-codepipeline-cfn-artifacts.html#w2ab1c21c15c15")

    parent_parser.add_argument('--profile', help='The named profile to use for AWS API calls.')

    parent_parser.add_argument('--enable-logging', help='Enable detailed logging.', default=False,
                               action='store_true')

    parent_parser.add_argument('--allow-dynamic-ref-without-version', help='Allow the retrieval of dynamic references without specifying a version number.  Specifying a version number'
                                ' helps ensure that the reference does not change between validation and deployment. Allowing dynamic references without versions would make it'
                                ' possible for the template you deploy to be different from the one that was validated.', default=False, action='store_true')

    parser = argparse.ArgumentParser(description='Parses IAM identity-based and resource-based policies from AWS CloudFormation templates.')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    subparsers = parser.add_subparsers(dest='{parse,validate}')
    subparsers.required = True

    # parse command
    parse_parser = subparsers.add_parser('parse', help='Replaces intrinsic and pseudo functions in an AWS CloudFormation '
                                                       'template and returns a JSON formatted list of IAM policies and '
                                                       'their attached resources.', parents=[parent_parser])
    parse_parser.set_defaults(func=parse_from_cli)

    # validate command
    validate_parser = subparsers.add_parser('validate', help='Parses IAM identity-based and resource-based policies from AWS CloudFormation templates '
                                                           'and runs them through IAM Access Analyzer for validation.  Returns the findings from '
                                                           'validation in JSON format.', parents=[parent_parser])
    validate_parser.set_defaults(func=validate_from_cli)

    validate_parser.add_argument('--ignore-finding', dest="ignore_finding", metavar='FINDING_CODE,RESOURCE_NAME,RESOURCE_NAME.FINDING_CODE',
                                 help='Allow validation failures to be ignored.\n'
                             'Specify as a comma separated list of findings to be ignored. Can be individual '
                             'finding codes (e.g. "PASS_ROLE_WITH_STAR_IN_RESOURCE"), a specific resource name '
                             '(e.g. "MyResource"), or a combination of both separated by a period.'
                             '(e.g. "MyResource.PASS_ROLE_WITH_STAR_IN_RESOURCE").',
                                 action=ParseFindingsToIgnoreFromCLI)

    validate_parser.add_argument('--treat-finding-type-as-blocking', dest="treat_as_blocking", metavar="ERROR,SECURITY_WARNING",
                                 help='Specify which finding types should be treated as blocking. Other finding types are treated '
                                     'as non-blocking. Defaults to "ERROR" and "SECURITY_WARNING". Specify as a comma separated '
                                     'list of finding types that should be blocking.  Possible values are "ERROR", '
                                     '"SECURITY_WARNING", "SUGGESTION", and "WARNING".  Pass "NONE" to ignore all errors.',
                                 default=default_finding_types_that_are_blocking, type=validate_finding_types_from_cli)

    validate_parser.add_argument('--allow-external-principals', dest='allowed_external_principals', metavar="ACCOUNT,ARN",
                                 help='A comma separated list of external principals that should be ignored.  Specify as '
                                     'a comma separated list of a 12 digit AWS account ID, a federated web identity '
                                     'user, a federated SAML user, or an ARN. Specify "*" to allow anonymous access. '
                                     '(e.g. 123456789123,arn:aws:iam::111111111111:role/MyOtherRole,graph.facebook.com)',
                                 action=ParseAllowExternalPrincipalsFromCLI)

    args = parser.parse_args(args)

    try:
        client.set_profile(args.profile)
        validate_credentials(args.region)

        configure_logging(args.enable_logging)
        args.parameters = parameters.merge(args.parameters, args.template_configuration_file)
        args.func(args)
    except ApplicationError as e:
        print(f'ERROR: {str(e)}', file=sys.stderr)
        exit(1)
    except Exception as e:
        traceback.print_exc()
        print(f'ERROR: Unexpected error occurred. {str(e)}', file=sys.stderr)
        exit(1)


if __name__ == "__main__":
    main()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.cfn_tools.common_schema import array_of_strings_schema

# schema for role/group/user Policies property
policies_schema = {
	'type': 'array',
	'items': {
		'type': 'object',
		'properties': {
			'PolicyDocument': {
				'type': 'object'
			},
			'PolicyName': {
				'type': 'string'
			}
		},
		'required': ['PolicyDocument', 'PolicyName']
	}
}

managed_policy_arns_schema = {
	'type': 'array',
	'items': {
		'type': 'string'
	}
}


groups_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'GroupName': {
					'type': 'string'
				},
				'Policies': policies_schema,
				'ManagedPolicyArns': managed_policy_arns_schema
			}
		}
	}
}

roles_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'RoleName': {
					'type': 'string',
					"pattern": r'^([\w+=,.@-]+)$'
				},
				'AssumeRolePolicyDocument': {
					'type': 'object'
				},
				'Policies': policies_schema,
				'ManagedPolicyArns': managed_policy_arns_schema
			},
			'required': ['AssumeRolePolicyDocument']
		}
	},
	'required': ['Properties']
}

users_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'UserName': {
					'type': 'string'
				},
				'Policies': policies_schema,
				'ManagedPolicyArns': managed_policy_arns_schema
			}
		}
	}
}

permission_set_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Name': {
					'type': 'string'
				},
				'InlinePolicy': {
					'type': 'object'
				},
				'ManagedPolicies': managed_policy_arns_schema
			},
			'required': ['Name']
		}
	},
	'required': ['Properties']
}

managed_policy_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'ManagedPolicyName': {
					'type': 'string'
				},
				'PolicyDocument': {
					'type': 'object'
				},
				'Path': {
					'type': 'string'
				},
				'Roles': array_of_strings_schema,
				'Users': array_of_strings_schema,
				'Groups': array_of_strings_schema
			},
			'required': ['PolicyDocument']
		}
	},
	'required': ['Properties']
}

inline_policy_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'PolicyName': {
					'type': 'string'
				},
				'PolicyDocument': {
					'type': 'object'
				},
				'Roles': array_of_strings_schema,
				'Users': array_of_strings_schema,
				'Groups': array_of_strings_schema
			},
			'required': ['PolicyName', 'PolicyDocument']
		}
	},
	'required': ['Properties']
}


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.cfn_tools.cfn_object import CfnObject


class NoValue:
	"""
	This is a placeholder used to replace a reference to NoValue - {"Ref": "NoValue"}.  These placeholders (and dict keys
	that reference them) are removed from the evaluated output after resource evaluation.
	"""
	pass


class AwsNoValueEvaluator:
	"""
	Removes any references to NoValue from the evaluated output.  If a dictionary key has a NoValue reference, the key
	is also removed.
	"""
	def evaluate(self, value):
		if isinstance(value, dict):
			for key in list(value.keys()):
				child_value = value[key]
				if isinstance(child_value, NoValue):
					del value[key]
				else:
					self.evaluate(child_value)

		elif isinstance(value, list):
			for item in value:
				self.evaluate(item)

			# mutate the list to remove any item that is NoValue
			# must be done outside of the above for loop
			value[:] = [item for item in value if not isinstance(item, NoValue)]

		return value



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import argparse

from cfn_policy_validator.validation.reporter import ResourceOrCodeFindingToIgnore, ResourceAndCodeFindingToIgnore, \
    AllowedExternalArn, AllowedExternalPrincipal
from cfn_policy_validator.cfn_tools import regex_patterns


class DictionaryArgument(argparse.Action):
    """
    Converts key/value pairs in the format of Key=Value to a dictionary
    """

    def __call__(self, _, namespace, values, option_string=None):
        dictionary = {}
        for key_and_value in values:
            key, value = key_and_value.split("=", 1)
            dictionary[key] = value
        setattr(namespace, self.dest, dictionary)


class ParseFindingsToIgnoreFromCLI(argparse.Action):
    """
    Parses comma delimited list of findings to be ignored.  This is either a resource name or a finding code name, or
    a combination of both in the form MyResource.FindingA
    """

    def __call__(self,  _, namespace, values, option_string=None):
        values = values.split(',')

        findings_to_ignore = parse_findings_to_ignore(values)

        setattr(namespace, self.dest, findings_to_ignore)


def parse_findings_to_ignore(values_as_list):
    if values_as_list is None:
        return values_as_list

    values_as_list = [value.strip() for value in values_as_list]

    findings_to_ignore = []
    for value in values_as_list:
        if "." in value:
            resource_and_code = value.split(".", 1)
            # a split must have at least two members of the array, so no need to validate
            finding_to_ignore = ResourceAndCodeFindingToIgnore(resource_and_code[0], resource_and_code[1])
        else:
            finding_to_ignore = ResourceOrCodeFindingToIgnore(value)

        findings_to_ignore.append(finding_to_ignore)

    return findings_to_ignore


class ParseAllowExternalPrincipalsFromCLI(argparse.Action):
    """
    Parse a comma delimieted list of external principals that are allowed.  These are either principal ARNs or values
    (account IDs, canonical user, etc)
    """

    def __call__(self, _, namespace, values, option_string=None):
        values = values.split(',')

        allowed_external_principals = parse_allow_external_principals(values)

        setattr(namespace, self.dest, allowed_external_principals)


def parse_allow_external_principals(values_as_list):
    if values_as_list is None:
        return values_as_list

    values_as_list = [value.strip() for value in values_as_list]

    allowed_external_principals = []
    for value in values_as_list:
        match = regex_patterns.generic_arn_pattern.match(value)
        if match is None:
            allowed_external_principal = AllowedExternalPrincipal(value)
        else:
            allowed_external_principal = AllowedExternalArn(value)

        allowed_external_principals.append(allowed_external_principal)

    return allowed_external_principals



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from typing import TextIO

import json
import yaml

from cfn_policy_validator.cfn_tools import schema_validator
from cfn_policy_validator.cfn_tools.cfn_object import CfnObject
from cfn_policy_validator.cfn_tools.yaml_loader import CfnYamlLoader
from cfn_policy_validator.parsers.account_config import AccountConfig
from cfn_policy_validator.parsers.utils.node_evaluator import NodeEvaluator


def load(stream: TextIO, account_config: AccountConfig, parameters: dict, allow_dynamic_ref_without_version: bool):
	raw_template = stream.read()
	try:
		template = json.loads(raw_template, object_hook=CfnObject)
	except ValueError:
		template = yaml.load(raw_template, Loader=CfnYamlLoader)

	schema_validator.validate(template)
	node_evaluator = NodeEvaluator(template, account_config, parameters, allow_dynamic_ref_without_version)
	_populate_cfn_object(template, node_evaluator, [])
	return template


def _populate_cfn_object(d, node_evaluator, ancestors, parent=None):
	# recurse into a custom data structure that represents a node in the template
	if isinstance(d, CfnObject):
		d.add_ancestors(ancestors, parent)
		d.set_node_evaluator(node_evaluator)
		for key, value in d.items():
			_populate_cfn_object(value, node_evaluator, d.ancestors, key)
	elif isinstance(d, list):
		for item in d:
			_populate_cfn_object(item, node_evaluator, ancestors, parent)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator import ApplicationError
from cfn_policy_validator.parsers.output import Policy, Resource
from cfn_policy_validator.parsers.resource import get_parser_of_type


class S3BucketPolicyParser:
    """ AWS::S3::BucketPolicy
    """

    def __init__(self):
        self.bucket_policies = []

    def parse(self, _, resource):
        evaluated_resource = resource.eval(bucket_policy_schema)
        properties = evaluated_resource['Properties']

        bucket_name = properties['Bucket']
        policy_document = properties['PolicyDocument']

        policy = Policy('BucketPolicy', policy_document)
        resource = Resource(bucket_name, 'AWS::S3::Bucket', policy)

        self.bucket_policies.append(resource)

    def get_policies(self):
        return self.bucket_policies


bucket_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'Bucket': {
                    'type': 'string'
                },
                'PolicyDocument': {
                    'type': 'object'
                }
            },
            'required': ['Bucket', 'PolicyDocument']
        }
    },
    'required': ['Properties']
}


class S3BucketAclParser:
    """ AWS::S3::Bucket
    """
    def __init__(self):
        self.bucket_acls = []

    def parse(self, resource_name, resource):
        evaluated_resource = resource.eval(bucket_schema)
        properties = evaluated_resource.get('Properties', {})

        access_control = properties.get('AccessControl')
        if access_control is None:
            return

        bucket_name = properties.get('BucketName', resource_name)

        policy = Policy('BucketAcl', None)
        resource = Resource(bucket_name, 'AWS::S3::Bucket', policy, configuration={'AccessControl': access_control})

        self.bucket_acls.append(resource)

    def get_policies(self):
        return self.bucket_acls

    def merge_resource(self, bucket_acl, bucket_policy):
        bucket_policy.Configuration = bucket_acl.Configuration

    def merge_policies(self, other_resource_parsers):
        # look and see if there were any s3 bucket policies parsed
        s3_bucket_policy_parser = get_parser_of_type(other_resource_parsers, S3BucketPolicyParser)
        if s3_bucket_policy_parser is None:
            return

        unmerged_bucket_acls = []

        bucket_policies = s3_bucket_policy_parser.get_policies()
        for bucket_acl in self.bucket_acls:
            matching_bucket_policy = next((bucket_policy for bucket_policy in bucket_policies
                                        if bucket_acl.ResourceName == bucket_policy.ResourceName), None)

            # if there were no matching bucket policies, there is nothing to merge
            if matching_bucket_policy is None:
                unmerged_bucket_acls.append(bucket_acl)
            else:
                # there must be exactly 1 matching bucket policy at this point and we merge the ACL and bucket
                # policy into a single resource
                self.merge_resource(bucket_acl, matching_bucket_policy)

        self.bucket_acls = unmerged_bucket_acls


bucket_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'BucketName': {
                    'type': 'string'
                },
                'AccessControl': {
                    'type': 'string'
                }
            }
        }
    }
}


class S3AccessPointPolicyParser:
    """ AWS::S3::AccessPoint
    """

    def __init__(self):
        self.access_point_policies = []

    def parse(self, resource_name, resource):
        evaluated_resource = resource.eval(access_point_policy_schema)
        properties = evaluated_resource['Properties']

        policy_document = properties.get('Policy')
        if policy_document is None:
            # we don't need to parse access points that don't have policies and policy is optional
            return

        access_point_name = properties.get('Name', resource_name)
        vpc_id = properties.get('VpcConfiguration', {}).get('VpcId')

        configuration = None
        if vpc_id is not None:
            configuration = {
                'VpcId': vpc_id
            }

        policy = Policy('AccessPointPolicy', policy_document)
        resource = Resource(access_point_name, 'AWS::S3::AccessPoint', policy, configuration)

        self.access_point_policies.append(resource)

    def get_policies(self):
        return self.access_point_policies


access_point_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'Name': {
                    'type': 'string'
                },
                'Policy': {
                    'type': 'object'
                },
                'VpcConfiguration': {
                    'type': 'object'
                }
            }
        }
    },
    'required': ['Properties']
}


class S3MultiRegionAccessPointPolicyParser:
    """ AWS::S3::MultiRegionAccessPointPolicy
    """

    def __init__(self):
        self.multi_region_access_point_policies = []

    def parse(self, _, resource):
        evaluated_resource = resource.eval(multi_region_access_point_policy_schema)
        properties = evaluated_resource['Properties']

        mrap_name = properties['MrapName']
        policy = properties['Policy']

        policy_document = Policy('MultiRegionAccessPointPolicy', policy)
        resource = Resource(mrap_name, 'AWS::S3::MultiRegionAccessPoint', policy_document)

        self.multi_region_access_point_policies.append(resource)

    def get_policies(self):
        return self.multi_region_access_point_policies


multi_region_access_point_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'MrapName': {
                    'type': 'string'
                },
                'Policy': {
                    'type': 'object'
                }
            },
            'required': ['MrapName', 'Policy']
        }
    },
    'required': ['Properties']
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
import os
import re

from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.parsers.utils.arn_generator_schemas import iam_role_schema, iam_user_schema, \
    elbv2_load_balancer_schema, elbv2_listener_schema, elbv2_target_group_schema, network_firewall_rulegroup_schema, \
    iam_managed_policy_schema


class ArnGenerator:
    """
    This class uses the cfn_to_arn_map.json file to automatically generate valid ARNs from a CloudFormation resource
    type.  It does this by replacing variables from the arn patterns in the cfn_to_arn_map.json file.  An arn pattern
    looks like this: arn:${Partition}:sqs:${Region}:${Account}:${QueueName}

    Any variable within ${..} and not named Partition, Region, or Account is replaced with the resource name.  This
    works because IAM policy validation does not rely on the name of the resource.
    """

    def __init__(self, account_config):
        this_files_directory = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(this_files_directory, 'cfn_to_arn_map.json')) as f:
            self.cfn_to_arn_map = json.load(f)

        self.account_config = account_config

        # some ARN generation requires custom logic (e.g. an ELB can have 2 different ARNs depending on the ELB type)
        self.custom_generators = {
            'AWS::ElasticLoadBalancingV2::LoadBalancer': generate_elbv2_load_balancer_arn,
            'AWS::ElasticLoadBalancingV2::Listener': generate_elbv2_listener_arn,
            'AWS::ElasticLoadBalancingV2::TargetGroup': generate_elbv2_target_group_arn,
            'AWS::IAM::Role': generate_role_arn,
            'AWS::IAM::User': generate_user_arn,
            'AWS::IAM::ManagedPolicy': generate_managed_policy_arn,
            'AWS::NetworkFirewall::RuleGroup': generate_network_firewall_rule_group
        }

    def try_generate_arn(self, resource_name, resource, attribute_or_ref, visited_values=None):
        if visited_values is None:
            visited_values = []

        cfn_type = resource['Type']
        split_cfn_type = cfn_type.split("::")
        if len(split_cfn_type) != 3:
            if len(split_cfn_type) == 4 and split_cfn_type[3].lower() == 'module':
                raise ApplicationError(f'Unable to resolve {cfn_type}. CloudFormation modules are not yet supported.')

            raise ApplicationError(f'Invalid resource type: {cfn_type}')

        service_name = split_cfn_type[1]
        resource_type = split_cfn_type[2]

        arn_pattern = self.cfn_to_arn_map\
            .get(service_name, {})\
            .get(resource_type, {})\
            .get(attribute_or_ref)

        if arn_pattern is None:
            return None

        arn_pattern = arn_pattern['Value']

        # arn_patterns follow the format: arn:${Partition}:sqs:${Region}:${Account}:${QueueName}
        # where Partition, Account, and Region are standardized variable names
        arn_pattern = arn_pattern.replace("${Partition}", self.account_config.partition)
        arn_pattern = arn_pattern.replace("${Region}", self.account_config.region)
        arn_pattern = arn_pattern.replace("${Account}", self.account_config.account_id)

        # certain CFN types require some additional generation that is specific to the resource type
        # for example, we include the exact path with any roles or users, ALBs and NLBs share the same cfn resource,
        # but have different ARNs
        custom_generator = self.custom_generators.get(cfn_type)
        if custom_generator is not None:
            arn_pattern = custom_generator(arn_pattern, resource_name, resource, visited_values)

        # match any variable (anything not Partition, Account, Region) within brackets and replace with the resource name
        # e.g. arn:aws:..:${SomeResourceName} -> arn:aws:..:ResourceName
        regex = r"(?P<Variable>\${[^}]*})"

        def callback(match):
            return resource_name

        arn_pattern = re.sub(regex, callback, arn_pattern)

        return arn_pattern


# AWS::IAM::Role
# include the path in the Role ARN which will be helpful for analysis
def generate_role_arn(arn_pattern, resource_name, resource, visited_values):
    evaluated_resource = resource.eval(iam_role_schema, visited_values)

    properties = evaluated_resource['Properties']

    path = properties.get('Path', '/')
    # remove the leading / to fit the template
    path = path[1:]

    name = properties.get('RoleName', resource_name)

    return arn_pattern.replace("${RoleNameWithPath}", path + name)


# AWS::IAM::User
# include the path in the User ARN which will be helpful for analysis
def generate_user_arn(arn_pattern, resource_name, resource, visited_values):
    evaluated_resource = resource.eval(iam_user_schema, visited_values)

    properties = evaluated_resource.get('Properties', {})

    path = properties.get('Path', '/')
    # remove the leading / to fit the template
    path = path[1:]

    name = properties.get('UserName', resource_name)
    return arn_pattern.replace("${UserNameWithPath}", path + name)


def generate_managed_policy_arn(arn_pattern, resource_name, resource, visited_values):
    evaluated_resource = resource.eval(iam_managed_policy_schema, visited_values)

    properties = evaluated_resource['Properties']

    path = properties.get('Path', '/')
    path = path[1:]

    name = properties.get('ManagedPolicyName', resource_name)
    return arn_pattern.replace("${PolicyNameWithPath}", path + name)


# Multiple load balancers share the same CFN resources, but have different ARNs depending on load balancer type
# AWS::ElasticLoadBalancingV2::LoadBalancer
def generate_elbv2_load_balancer_arn(arn_pattern, _, resource, visited_values):
    evaluated_resource = resource.eval(elbv2_load_balancer_schema, visited_values)
    properties = evaluated_resource.get('Properties', {})

    lb_type = properties.get('Type')
    if lb_type is None:
        lb_type = 'application'

    if lb_type == 'network':
        return arn_pattern.replace('loadbalancer/app/', 'loadbalancer/net/')
    elif lb_type == 'gateway':
        return arn_pattern.replace('loadbalancer/app/', 'loadbalancer/gwy/')

    return arn_pattern


# AWS::ElasticLoadBalancingV2::Listener
def generate_elbv2_listener_arn(arn_pattern, _, resource, visited_values):
    evaluated_resource = resource.eval(elbv2_listener_schema, visited_values)

    properties = evaluated_resource['Properties']
    protocol = properties.get('Protocol')

    if protocol is None:
        return arn_pattern.replace('listener/app/', 'listener/gwy/')

    if protocol not in ['HTTP', 'HTTPS']:
        return arn_pattern.replace('listener/app/', 'listener/net/')

    return arn_pattern


# AWS::ElasticLoadBalancingV2::TargetGroup
def generate_elbv2_target_group_arn(arn_pattern, _, resource, visited_values):
    evaluated_resource = resource.eval(elbv2_target_group_schema, visited_values)

    properties = evaluated_resource.get('Properties', {})
    protocol = properties.get('Protocol')

    if protocol is None:
        return arn_pattern

    if protocol == 'GENEVE':
        return arn_pattern.replace('loadbalancer/app/', 'loadbalancer/gwy/')

    if protocol in ['HTTP', 'HTTPS']:
        return arn_pattern

    return arn_pattern.replace('loadbalancer/app/', 'loadbalancer/net/')


# AWS::NetworkFirewall::RuleGroup
# network firewall rule groups can be either stateful or stateless and have different ARNs depending on the type
def generate_network_firewall_rule_group(arn_pattern, _, resource, visited_values):
    evaluated_resource = resource.eval(network_firewall_rulegroup_schema, visited_values)

    properties = evaluated_resource['Properties']
    rule_group_type = properties['Type']

    if rule_group_type == 'STATEFUL':
        return arn_pattern

    return arn_pattern.replace(':stateful-rulegroup/', ':stateless-rulegroup/')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.parsers.output import Policy, Resource


class KmsKeyPolicyParser:
    """ AWS::KMS::Key
    """

    def __init__(self):
        self.key_policies = []

    def parse(self, resource_name, resource):
        evaluated_resource = resource.eval(key_schema)
        properties = evaluated_resource['Properties']

        policy_document = properties['KeyPolicy']

        policy = Policy('KeyPolicy', policy_document)
        resource = Resource(resource_name, 'AWS::KMS::Key', policy)

        self.key_policies.append(resource)

    def get_policies(self):
        return self.key_policies


key_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'KeyPolicy': {
                    'type': 'object'
                }
            },
            'required': ['KeyPolicy']
        }
    },
    'required': ['Properties']
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
iam_role_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'RoleName': {
					'type': 'string'
				}
			}
		}
	},
	'required': ['Properties']
}

iam_user_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'UserName': {
					'type': 'string'
				}
			}
		}
	}
}

iam_managed_policy_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Path': {
					'type': 'string'
				},
				'ManagedPolicyName': {
					'type': 'string'
				}
			}
		}
	},
	'required': ['Properties']
}


elbv2_load_balancer_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Type': {
					'type': 'string'
				}
			}
		}
	}
}

elbv2_listener_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Protocol': {
					'type': 'string'
				}
			}
		}
	},
	'required': ['Properties']
}

elbv2_target_group_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Protocol': {
					'type': 'string'
				}
			}
		}
	}
}

network_firewall_rulegroup_schema = {
	'type': 'object',
	'properties': {
		'Properties': {
			'type': 'object',
			'properties': {
				'Type': {
					'type': 'string'
				}
			},
			'required': ['Type']
		}
	}
}


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfn_policy_validator import client

# only look up the user id if it's requested
canonical_user_id = None


# Resolution of the canonical user in an account which is a possible principal value for a policy and also
# used when evaluating S3 bucket ACLs.
def get_canonical_user(region):
    global canonical_user_id
    if canonical_user_id is not None:
        return canonical_user_id

    s3_client = client.build('s3', region)
    response = s3_client.list_buckets()
    canonical_user_id = response['Owner']['ID']
    return canonical_user_id



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema


class SelectEvaluator:
	def __init__(self, node_evaluator):
		self.node_evaluator = node_evaluator

	def evaluate(self, value, visited_values):
		validate_schema(value, select_schema, 'Fn::Select')

		try:
			index = self.node_evaluator.eval(value[0], visited_values=visited_values)
			index = int(index)
		except ValueError:
			raise ApplicationError(f'The first value for Fn::Select must be an integer. Invalid value: {value[0]}')

		list_of_objects = self.node_evaluator.eval_with_validation(value[1], list_of_objects_schema, path='Fn::Select.1', visited_values=visited_values)

		if index >= len(list_of_objects) or index < 0:
			raise ApplicationError(f'Fn::Select index is out of bounds of the list.  Invalid value: {value}')

		return self.node_evaluator.eval(list_of_objects[index], visited_values)


select_schema = {
	'type': 'array',
	'items': [
		# both index and list_of_objects can contain references
		{},
		{}
	],
	'additionalItems': False
}

index_schema = {
	'type': 'integer'
}

list_of_objects_schema = {
	'type': 'array'
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from urllib.parse import urlparse

from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.parsers.output import Policy, Resource


class SqsQueuePolicyParser:
    """ AWS::SQS::QueuePolicy
    """
    def __init__(self):
        self.queue_policies = []

    def parse(self, _, resource):
        evaluated_resource = resource.eval(sqs_queue_policy_schema)
        properties = evaluated_resource['Properties']

        queue_urls = properties['Queues']
        policy_document = properties['PolicyDocument']

        for queue in queue_urls:
            parsed_url = urlparse(queue)
            try:
                queue_name = parsed_url.path.split('/')[2]
            except IndexError:
                raise ApplicationError(f'Invalid queue URL. Unable to parse name from URL. Invalid value: "{queue}"')

            policy = Policy('QueuePolicy', policy_document)
            resource = Resource(queue_name, 'AWS::SQS::Queue', policy)

            self.queue_policies.append(resource)

    def get_policies(self):
        return self.queue_policies


sqs_queue_policy_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'PolicyDocument': {
                    'type': 'object'
                },
                'Queues': {
                    'type': 'array',
                    'minItems': 1,
                    'items': {
                        'type': 'string'
                    }
                }
            },
            'required': ['PolicyDocument', 'Queues']
        }
    },
    'required': ['Properties']
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.canonical_user_id import get_canonical_user
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema
from cfn_policy_validator.parsers.utils.cycle_detection import validate_no_cycle
from cfn_policy_validator.parsers.utils.intrinsic_functions import name_hints


def validate_fn_get_att_schema(value):
	validate_schema(value, get_att_schema, 'Fn::GetAtt')


class GetAttEvaluator:
	def __init__(self, resources, arn_generator, node_evaluator, region):
		self.resources = resources
		self.arn_generator = arn_generator
		self.node_evaluator = node_evaluator
		self.region = region
		self.custom_get_att_evals = {
			'AWS::CloudFront::CloudFrontOriginAccessIdentity': {
				'S3CanonicalUserId': get_canonical_user
			}
		}

	def evaluate(self, get_att_lookup, visited_values=None):
		if visited_values is None:
			visited_values = []

		validate_fn_get_att_schema(get_att_lookup)

		logical_name_of_resource = get_att_lookup[0]
		attribute_name = get_att_lookup[1]

		resource = self.resources.get(logical_name_of_resource)
		if resource is None:
			raise ApplicationError(f'Unable to find referenced resource for GetAtt reference to {logical_name_of_resource}.{attribute_name}')

		resource_type = resource['Type']
		properties = resource.get('Properties', {})

		explicit_name_property = name_hints.get(resource_type)
		explicit_resource_name = properties.get(explicit_name_property)
		if explicit_resource_name is None:
			# if an explicit name is not specified, default to the logical name
			resource_name = logical_name_of_resource
		else:
			# we found a valid property value for the name of the resources. this property may reference another resource,
			# so check to see if we've already done that and we're stuck in a cycle
			validate_no_cycle(logical_name_of_resource, explicit_name_property, visited_values)
			resource_name = self.node_evaluator.eval(explicit_resource_name, visited_values=visited_values)

		arn = self.arn_generator.try_generate_arn(resource_name, resource, attribute_name, visited_values=visited_values)
		if arn is not None:
			return arn

		# if the GetAtt does not reference an ARN, see if we have a custom evaluation for this get_att.
		# Useful in cases where an attribute returns something other than an ARN that's relevant to
		# IAM policies (canonical username)
		custom_get_att_eval = self.custom_get_att_evals.get(resource['Type'], {}).get(attribute_name)
		if custom_get_att_eval is not None:
			return custom_get_att_eval(self.region)

		# For calls to GetAtt that are not ARNs, try to find a property with the same name.  This is a last resort and
		# should probably not occur often.  We expect to almost always see GetAtt used for ARNs in the context of an
		# IAM policy
		properties = resource.get('Properties', {})
		property_value = properties.get(attribute_name)
		if property_value is None:
			raise ApplicationError(f'Call to GetAtt not supported for: {logical_name_of_resource}.{attribute_name}')

		# we may need to traverse to another resource, so check to see if we've already done that and
		# we're stuck in a cycle
		validate_no_cycle(logical_name_of_resource, attribute_name, visited_values)

		# there are many return types for GetAtt, so it's the caller's responsibility to validate expected type
		return self.node_evaluator.eval(property_value, visited_values=visited_values)


get_att_schema = {
	'type': 'array',
	'items': [
		# only attribute name can contain references
		{'type': 'string'},
		{}
	],
	'additionalItems': False
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from json import JSONDecodeError

from argparse import ArgumentTypeError
from botocore.config import Config
from botocore.exceptions import InvalidRegionError
from botocore.utils import validate_region_name

from cfn_policy_validator import client
from cfn_policy_validator.application_error import ApplicationError


def merge(parameters, template_configuration_file_path):
	"""
	Merge parameters passed in via the parameters argument with parameters from the template configuration file.
	"""

	template_configuration_file = {}
	if template_configuration_file_path is not None:
		template_configuration_file = _read_template_configuration_file(template_configuration_file_path)

	parameters_from_config_file = template_configuration_file.get('Parameters', {})
	if not isinstance(parameters_from_config_file, dict):
		raise ApplicationError(f'The value for "Parameters" in the template configuration value must be a JSON object.\n'
								'See CloudFormation documentation on format for this file: '
								'"https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/continuous-delivery-codepipeline-cfn-artifacts.html#w2ab1c21c15c15"')

	if parameters is None:
		parameters = {}

	# parameters passed on the command line take precedence over those in the configuration file
	# this is because the command line will typically be more controlled / trusted than the config file
	parameters_from_config_file.update(parameters)

	return parameters_from_config_file


def _read_template_configuration_file(template_configuration_file_path):
	try:
		with open(template_configuration_file_path, 'r') as stream:
			raw_file = stream.read()
			return json.loads(raw_file)
	except FileNotFoundError:
		raise ApplicationError(f'Template configuration file not found: {template_configuration_file_path}')
	except JSONDecodeError:
		raise ApplicationError(f'Template configuration file contains invalid json: {template_configuration_file_path}')


def validate_region(region):
	try:
		# this call validates that the region name is valid, but does not validate that the region actually exists
		validate_region_name(region)
	except InvalidRegionError:
		raise ArgumentTypeError(f'Invalid region name: {region}.')

	return region


def validate_credentials(region):
	# run a test to validate the provided credentials
	# create our own config here to control retries and fail fast if credentials are invalid
	sts_client = client.build('sts', region, client_config=Config(retries={'mode': 'standard', 'max_attempts': 2}))
	sts_client.get_caller_identity()


def validate_finding_types_from_cli(value):
	"""
	Validate that the finding types provided are valid finding types.
	"""

	finding_types = value.split(',')
	finding_types = validate_finding_types(finding_types)

	return finding_types


def validate_finding_types(finding_types):
	if finding_types is None:
		return finding_types

	finding_types = [finding_type.strip() for finding_type in finding_types]
	finding_types = [finding_type.upper() for finding_type in finding_types]

	for finding_type in finding_types:
		if finding_type not in ['ERROR', 'SECURITY_WARNING', 'SUGGESTION', 'WARNING', 'NONE']:
			raise ArgumentTypeError(f"Invalid finding type: {finding_type}.")

	return finding_types



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfn_policy_validator.application_error import ApplicationError


def validate_no_cycle(resource_name, resource_property, visited_values):
    """
    Validate that there are no recursive cycles present in CFN intrinsic functions.
    """

    node_to_check = VisitedNode(resource_name, resource_property)
    # if we've seen this node before, there must be a cycle
    if node_to_check in visited_values:
        raise ApplicationError(f'Cycle detected for {resource_name} and {resource_property}.')

    visited_values.append(node_to_check)


class VisitedNode:
    def __init__(self, resource_name, property_name):
        self.resource_name = resource_name
        self.property_name = property_name

    def __eq__(self, other):
        return isinstance(other, type(self)) \
            and (self.resource_name, self.property_name) == \
                (other.resource_name, other.property_name)

    def __hash__(self):
        return hash((self.resource_name, self.property_name))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from typing import Any


class CfnObject(dict):
	"""
	A representation of a JSON/YAML object that is aware of some context of where it exists in the template. Simplifies
	evaluations, allows for clearer error messages on failures, and avoids passing the node_evaluator around everywhere.
	"""

	def __init__(self, value={}):
		super(CfnObject, self).__init__(value)
		self.ancestors = []
		self.node_evaluator = None

	def add_ancestors(self, parent_ancestors: list, parent_key: str):
		self.ancestors = list(parent_ancestors)
		if parent_key is not None and parent_key != 'Resources':
			self.ancestors.append(parent_key)

	def set_node_evaluator(self, node_evaluator):
		self.node_evaluator = node_evaluator

	@property
	def parent(self):
		parent = self.ancestors[-1:]
		if len(parent) == 0:
			return None

		return parent[0]

	def eval(self, expected_schema: Any, visited_values: list = None):
		path = self.ancestors_as_string()

		resource_properties = self._find_resource_properties_in_schema(expected_schema)
		if resource_properties is not None:
			resource_properties = list(resource_properties.keys())

		return self.node_evaluator.eval_with_validation(self, expected_schema, resource_properties_to_eval=resource_properties, path=path, visited_values=visited_values)

	# iterate through schema and find the properties of the resource.  When we evaluate the resource, we only evaluate
	# the properties that we need to parse the policies
	def _find_resource_properties_in_schema(self, schema):
		properties = schema.get('properties')
		if properties is None:
			return None

		for key, value in properties.items():
			if key == 'Properties':
				return properties[key].get('properties')

			return self._find_resource_properties_in_schema(properties[key])

	def ancestors_as_string(self):
		if len(self.ancestors) == 0:
			return "Template"

		return ".".join(self.ancestors)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import codecs
import os
import setuptools
import re


def get_version(filename):
    with codecs.open(filename, 'r', 'utf-8') as fp:
        contents = fp.read()
    return re.search(r"__version__ = ['\"]([^'\"]+)['\"]", contents).group(1)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


version = get_version('cfn_policy_validator/version.py')

setuptools.setup(
    name="cfn-policy-validator",
    packages=setuptools.find_packages(exclude=["*.tests", "*_tests"]),
    version=version,
    author="matluttr",
    author_email="matluttr@amazon.com",
    description="Parses IAM identity-based and resource-based policies from a CloudFormation template and runs them through IAM Access Analyzer checks.",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    url='https://github.com/awslabs/aws-cloudformation-iam-policy-validator',
    keywords='cfn-policy-validator aws iam access-analyzer cloudformation',
    license='MIT-0',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    entry_points={"console_scripts": "cfn-policy-validator=cfn_policy_validator.main:main"},
    python_requires='>=3.6',
    package_data={
        '': ['*.json']
    },
    install_requires=[
        'boto3>=1.20',
        'pyYAML>=5.3',
        'urllib3>=1.25',
        'jsonschema~=3.2'
    ]
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from botocore.exceptions import ClientError

from cfn_policy_validator import client
from cfn_policy_validator.application_error import ApplicationError
from cfn_policy_validator.cfn_tools.regex_patterns import dynamic_ssm_reference_regex


class DynamicReferenceEvaluator:
	def __init__(self, region, allow_dynamic_ref_without_version):
		self.ssm_client = client.build('ssm', region)
		self.region = region
		self.allow_dynamic_ref_without_version = allow_dynamic_ref_without_version

	def evaluate(self, value):
		dynamic_ssm_references = dynamic_ssm_reference_regex.findall(value)
		for dynamic_ssm_reference in dynamic_ssm_references:
			dynamic_reference_text = dynamic_ssm_reference[0]
			parameter_name = dynamic_ssm_reference[1]
			parameter_version = dynamic_ssm_reference[2]

			if parameter_version == '':
				parameter = parameter_name
			else:
				parameter = f'{parameter_name}:{parameter_version}'

			if not self.allow_dynamic_ref_without_version and parameter_version == '':
				raise ApplicationError('Dynamic references to SSM parameters must include a version number to ensure the'
									   ' value does not change between validation and deployment. Use the '
									   '--allow-dynamic-ref-without-version flag if you want to disable this check for a version number. Invalid dynamic '
									   f'reference: {dynamic_reference_text}')

			try:
				response = self.ssm_client.get_parameter(Name=parameter)
			except ClientError as e:
				if e.response['Error']['Code'] == 'ParameterNotFound':
					raise ApplicationError(f'Could not find SSM parameter referenced by dynamic reference in region {self.region}: {dynamic_reference_text}')
				if e.response['Error']['Code'] == 'ParameterVersionNotFound':
					raise ApplicationError(f'Could not find version {parameter_version} of SSM parameter referenced by dynamic reference: {dynamic_reference_text}')
				raise

			value = value.replace(dynamic_ssm_reference[0], response['Parameter']['Value'])

		return value



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import itertools

from collections import defaultdict
from cfn_policy_validator.parsers.output import Policy, Resource


class LambdaPermissionPolicyParser:
    """ AWS::Lambda::Permission
    """
    def __init__(self, account_config):
        self.partition = account_config.partition
        self.region = account_config.region
        self.account_id = account_config.account_id

        self.permissions_policies = defaultdict(list)

    def parse(self, _, resource):
        evaluated_resource = resource.eval(lambda_permission_schema)

        properties = evaluated_resource['Properties']
        action = properties['Action']
        function_name = properties['FunctionName']
        principal_name = properties['Principal']

        source_account = properties.get('SourceAccount')
        source_arn = properties.get('SourceArn')

        policy_statement = {
            'Effect': 'Allow',
            'Action': action
        }

        if 'amazonaws' in principal_name:
            # this is a permission to a service
            policy_statement['Principal'] = {'Service': principal_name}
        else:
            # otherwise this is a permission to an account
            policy_statement['Principal'] = {'AWS': principal_name}

        # function name can be a full ARN, a partial ARN (a full suffix of an ARN) or the actual function name
        if function_name.startswith('arn:'):  # full ARN
            policy_statement['Resource'] = function_name
        elif 'function:' in function_name:  # partial ARN
            # partial arn must include at least function: and ":" is not a valid function name character
            arn_parts = function_name.split(":")
            # grab all parts of the partial ARN up until "function:"
            parts_before_function = list(itertools.takewhile(lambda part: part.lower() != 'function', arn_parts))

            # a list of a valid ARN before "function:"
            valid_arn_parts_before_function = ["arn", self.partition, "lambda", self.region, self.account_id]

            # construct the base_arn based on what's missing in the partial ARN
            number_of_missing_arn_parts = len(valid_arn_parts_before_function) - len(parts_before_function)
            base_arn = ":".join(valid_arn_parts_before_function[:number_of_missing_arn_parts])

            policy_statement['Resource'] = f'{base_arn}:{function_name}'
        else:  # just the function name
            policy_statement['Resource'] = f'arn:{self.partition}:lambda:{self.region}:{self.account_id}:function:{function_name}'

        if source_account is not None:
            policy_statement['Condition'] = {'StringEquals': {'AWS:SourceAccount': source_account}}

        if source_arn is not None:
            condition = policy_statement.get('Condition', {})
            condition['ArnLike'] = {'AWS:SourceArn': source_arn}
            policy_statement['Condition'] = condition

        # The permissions policies dictionary holds the statements per lambda function.  The statements are built
        # into an actual policy when they are retrieved.
        function_arn = policy_statement['Resource']
        statements = self.permissions_policies[function_arn]
        statements.append(policy_statement)

    def get_policies(self):
        resources = []
        for function_arn, statements in self.permissions_policies.items():
            function_name = self.__get_function_name_from_arn(function_arn)

            policy_document = {
                'Version': '2012-10-17',
                'Statement': statements
            }
            policy = Policy('PermissionsPolicy', policy_document)
            resource = Resource(function_name, "AWS::Lambda::Function", policy)
            resources.append(resource)

        return resources

    @staticmethod
    def __get_function_name_from_arn(arn_or_partial_arn):
        split_arn = arn_or_partial_arn.split(':')
        index_of_function = split_arn.index('function')
        return split_arn[index_of_function + 1]


lambda_permission_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'Action': {
                    'type': 'string'
                },
                'FunctionName': {
                    'type': 'string'
                },
                'Principal': {
                    'type': 'string'
                },
                'SourceAccount': {
                    'type': 'string'
                },
                'SourceArn': {
                    'type': 'string'
                }
            },
            'required': ['Action', 'FunctionName', 'Principal']
        }
    },
    'required': ['Properties']
}


class LambdaLayerVersionPermissionParser:
    def __init__(self, partition):
        self.partition = partition
        self.permissions_policies = defaultdict(list)

    def parse(self, _, resource):
        evaulated_resources = resource.eval(lambda_layer_version_permission_schema)

        properties = evaulated_resources['Properties']

        action = properties['Action']
        layer_version_arn = properties['LayerVersionArn']
        principal = properties['Principal']
        organization_id = properties.get('OrganizationId')

        policy_statement = {
            'Effect': 'Allow',
            'Action': action,
            'Resource': layer_version_arn,
            'Principal': {'AWS': principal}
        }

        if organization_id is not None:
            policy_statement['Condition'] = {'StringEquals': {'aws:PrincipalOrgID': organization_id}}

        # The permissions policies dictionary holds the statements per lambda layer.  The statements are built
        # into an actual policy when they are retrieved.
        statements = self.permissions_policies[layer_version_arn]
        statements.append(policy_statement)

    def get_policies(self):
        resources = []
        for layer_version_arn, statements in self.permissions_policies.items():
            layer_name = self.__get_layer_name_from_arn(layer_version_arn)
            policy_document = {
                'Version': '2012-10-17',
                'Statement': statements
            }
            policy = Policy('LayerVersionPermission', policy_document)
            resource = Resource(layer_name, "AWS::Lambda::LayerVersion", policy)
            resources.append(resource)

        return resources

    @staticmethod
    def __get_layer_name_from_arn(arn_or_partial_arn):
        split_arn = arn_or_partial_arn.split(':')
        index_of_function = split_arn.index('layer')
        name_and_version = split_arn[index_of_function + 1:]
        return ":".join(name_and_version)


lambda_layer_version_permission_schema = {
    'type': 'object',
    'properties': {
        'Properties': {
            'type': 'object',
            'properties': {
                'Action': {
                    'type': 'string'
                },
                'LayerVersionArn': {
                    'type': 'string'
                },
                'Principal': {
                    'type': 'string'
                },
                'OrganizationId': {
                    'type': 'string'
                }
            },
            'required': ['Action', 'LayerVersionArn', 'Principal']
        }
    },
    'required': ['Properties']
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re
from cfn_policy_validator.cfn_tools import regex_patterns
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_get_att_evaluator import validate_fn_get_att_schema
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_sub_evaluator import validate_fn_sub_schema


class TopologicalSorter:
    """
    Topologically sort CloudFormation resources so that we can process them in dependency order.  This allows us
    to simplify logic elsewhere by knowing that any dependencies were already processed.
    """

    def __init__(self, template):
        self.template = template
        self.resources = self.template['Resources']
        self.sorted_resources = []
        self.visited_resource_names = []

    def sort_resources(self):
        for resource_name in self.resources:
            if resource_name not in self.visited_resource_names:
                self.__inner_sort(resource_name, self.resources[resource_name])

        return self.sorted_resources

    def __inner_sort(self, resource_name, resource_value):
        self.visited_resource_names.append(resource_name)

        visitor = ResourceVisitor(resource_name, resource_value, self.resources)

        # we need to create a starting node for the crawler to perform a DFS on the tree and call ResourceVisitor
        # on each visited node
        parent = CloudFormationNode("Resources", self.resources)
        resource = CloudFormationNode(resource_name, resource_value, parent=parent, visitors=[visitor])
        resource.visit_children()

        node = visitor.get_node()

        for dependency_name in node.depends_on:
            if dependency_name not in self.visited_resource_names:
                dependent_resource_value = self.resources[dependency_name]
                self.__inner_sort(dependency_name, dependent_resource_value)

        self.sorted_resources.append(node)


class CloudFormationNode:
    def __init__(self, key, value, parent=None, ancestors=None, visitors=None):
        """
        A representation of a CloudFormation node that implements the visitor pattern
        """
        self.parent = parent

        self.ancestors = [] if ancestors is None else ancestors
        self.key = key
        self.value = value

        self.is_resource = self.parent is not None and self.parent.key == 'Resources'
        self.is_reference = self.key == "Ref"
        self.is_get_att = self.key == "Fn::GetAtt"
        self.is_sub = self.key == "Fn::Sub"

        self.__visitors = [] if visitors is None else visitors

    def visit_children(self):
        """ Visits all children of the current node and
        continues recursively for child dictionaries
        """

        for visitor in self.__visitors:
            visitor.visit(self)

        copy_of_ancestors = list(self.ancestors)
        copy_of_ancestors.append(self)

        if isinstance(self.value, dict):
            self._iterate_over_dict(self.value, copy_of_ancestors)
        elif isinstance(self.value, list):
            self._iterate_over_list(self.value, copy_of_ancestors)

    def _iterate_over_list(self, value_list, ancestors):
        for value in value_list:
            # the iterator treats items in a list as nodes with a key of None
            cfn_node = CloudFormationNode(None, value, self, ancestors, self.__visitors)
            cfn_node.visit_children()

    def _iterate_over_dict(self, dictionary, ancestors):
        for key in dictionary:
            value = dictionary[key]

            cfn_node = CloudFormationNode(key, value, self, ancestors, self.__visitors)
            cfn_node.visit_children()


class ResourceVisitor:
    def __init__(self, resource_name, resource_value, resources):
        self.dependencies = set()
        self.resource_name = resource_name
        self.resource_value = resource_value
        self.all_resources = resources

    def visit(self, cfn_node):
        if cfn_node.is_reference:
            self.__add_dependency_if_resource(cfn_node.value)

        elif cfn_node.is_get_att:
            validate_fn_get_att_schema(cfn_node.value)
            reference_name = cfn_node.value[0]
            self.__add_dependency_if_resource(reference_name)

        elif cfn_node.is_sub:
            validate_fn_sub_schema(cfn_node.value)

            if isinstance(cfn_node.value, list):
                string_to_evaluate = cfn_node.value[0]
                self.__add_fn_sub_dependencies(string_to_evaluate)
            else:
                self.__add_fn_sub_dependencies(cfn_node.value)

        elif cfn_node.is_resource:
            explicit_dependencies = cfn_node.value.get('DependsOn')
            if not isinstance(explicit_dependencies, list):
                explicit_dependencies = [explicit_dependencies]

            for dependency in explicit_dependencies:
                self.__add_dependency_if_resource(dependency)

    def __add_fn_sub_dependencies(self, value):
        matches = re.findall(regex_patterns.fn_sub_variables, value)
        for match in matches:
            if '.' in match:
                reference_name = match.split('.')[0]
                self.__add_dependency_if_resource(reference_name)
            else:
                self.__add_dependency_if_resource(match)

    def __add_dependency_if_resource(self, name):
        if name in self.all_resources:
            self.dependencies.add(name)

    def get_node(self):
        return Node(self.resource_name, self.resource_value, self.dependencies)


class Node:
    def __init__(self, logical_name, value, dependencies):
        self.logical_name = logical_name
        self.value = value
        self.depends_on = dependencies



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
import sys

LOGGER = logging.getLogger('cfn-policy-validator')


def configure_logging(enable_logging):
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)

    LOGGER.setLevel(logging.INFO)

    log_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(log_formatter)

    for handler in LOGGER.handlers:
        LOGGER.removeHandler(handler)
    LOGGER.addHandler(console_handler)

    if not enable_logging:
        LOGGER.disabled = True



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfn_policy_validator.application_error import SchemaValidationError
from cfn_policy_validator.cfn_tools import regex_patterns
from cfn_policy_validator.cfn_tools.cfn_object import CfnObject
from cfn_policy_validator.cfn_tools.common_schema import string_schema
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema


def validate_fn_sub_schema(value):
	try:
		validate_schema(value, sub_schema, 'Fn::Sub')
	except SchemaValidationError as e:
		if "is not of type 'array'" in str(e):
			# since GetAtt has multiple schemas, provide a custom error message that includes both schema options.
			# AnyOf just raises an arbitrary schema as an error.
			raise SchemaValidationError(f"{value} is not of type 'array or string', Path: Fn::Sub")
		else:
			raise


class SubEvaluator:
	def __init__(self, ref_evaluator, get_att_evaluator, node_evaluator):
		self.get_att_evaluator = get_att_evaluator
		self.ref_evaluator = ref_evaluator
		self.node_evaluator = node_evaluator

	def evaluate(self, value: CfnObject, visited_values: list = None):
		if visited_values is None:
			visited_values = []

		validate_fn_sub_schema(value)

		# There are two ways or writing a Fn::Sub function, a list and a string
		if isinstance(value, list):
			# a Fn::Sub with a list value has a variable map
			string_to_evaluate = value[0]
			variable_mapping = value[1]

			variables_in_string = re.findall(regex_patterns.fn_sub_variables, string_to_evaluate)
			for variable_in_string in variables_in_string:
				variable_value = variable_mapping.get(variable_in_string)
				if variable_value is not None:
					variable_value = self.node_evaluator.eval_with_validation(variable_value, string_schema, visited_values=visited_values)
					string_to_evaluate = string_to_evaluate.replace("${" + variable_in_string + "}", variable_value)
				else:
					# this will throw if it can't find the variable anywhere
					string_to_evaluate = self._evaluate_ref_or_get_att_in_sub(string_to_evaluate, variable_in_string, visited_values)

			return string_to_evaluate
		else:
			# the sub is a string value
			variables_in_string = re.findall(regex_patterns.fn_sub_variables, value)
			for variable_in_string in variables_in_string:
				value = self._evaluate_ref_or_get_att_in_sub(value, variable_in_string, visited_values)

			return value

	def _evaluate_ref_or_get_att_in_sub(self, value, variable_in_string, visited_values):
		if '.' in variable_in_string:
			# e.g. variable is MyResource.Arn
			variable_value = self.get_att_evaluator.evaluate(variable_in_string.split('.'), visited_values=visited_values)
		else:
			# e.g. variable is just MyResource
			variable_value = self.ref_evaluator.evaluate(variable_in_string, visited_values=visited_values)

		# validate the evaluated value is a string
		validate_schema(variable_value, string_schema, 'Fn::Sub')
		return value.replace("${" + variable_in_string + "}", variable_value)


sub_with_variable_map_schema = {
	"type": "array",
	"items": [
		{"type": "string"},
		{"type": "object"}
	],
	"additionalItems": False
}

sub_without_variable_map_schema = {
	"type": "string"
}


sub_schema = {
	"anyOf": [
		sub_with_variable_map_schema,
		sub_without_variable_map_schema
	]
}



# used to map common CFN resource types to their name properties to more accurately generate names for common resources
# instead of defaulting to the CFN resource's logical name
name_hints = {
	'AWS::S3::Bucket': 'BucketName',
	'AWS::Lambda::Function': 'FunctionName',
	'AWS::IAM::Role': 'RoleName',
	'AWS::IAM::User': 'UserName',
	'AWS::IAM::Group': 'GroupName',
	'AWS::SQS::Queue': 'QueueName',
	'AWS::SSM::Parameter': 'Name',
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import copy
import functools

from cfn_policy_validator.cfn_tools.cfn_object import CfnObject
from cfn_policy_validator.cfn_tools.schema_validator import validate_schema
from cfn_policy_validator.parsers.utils.arn_generator import ArnGenerator
from cfn_policy_validator.parsers.utils.intrinsic_functions.aws_no_value_evaluator import AwsNoValueEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.dynamic_ref_evaluator import DynamicReferenceEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_find_in_map_evaluator import FindInMapEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_get_att_evaluator import GetAttEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_import_value_evaluator import ImportValueEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_join_evaluator import JoinEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_select_evaluator import SelectEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_split_evaluator import SplitEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.fn_sub_evaluator import SubEvaluator
from cfn_policy_validator.parsers.utils.intrinsic_functions.ref_evaluator import RefEvaluator

from typing import Any


def evaluate_dynamic_references(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        dynamic_reference_evaluator = args[0].dynamic_reference_evaluator
        value = func(*args, **kwargs)
        if isinstance(value, str):
            return dynamic_reference_evaluator.evaluate(value)
        else:
            return value

    return wrapper


def prune_references_to_no_value(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        no_value_evaluator = args[0].no_value_evaluator
        value = func(*args, **kwargs)
        return no_value_evaluator.evaluate(value)

    return wrapper


class NodeEvaluator:
    def __init__(self, template, account_config, parameter_values, allow_dynamic_ref_without_version: bool):
        resources = template['Resources']
        parameters = template.get('Parameters', {})
        mappings = template.get('Mappings', {})

        arn_generator = ArnGenerator(account_config)

        self.dynamic_reference_evaluator = DynamicReferenceEvaluator(account_config.region, allow_dynamic_ref_without_version)
        self.no_value_evaluator = AwsNoValueEvaluator()

        ref_evaluator = RefEvaluator(resources, arn_generator, parameters, parameter_values, account_config, self)
        get_att_evaluator = GetAttEvaluator(resources, arn_generator, self, account_config.region)
        self.evaluators = {
            'Ref': ref_evaluator,
            'Fn::GetAtt': get_att_evaluator,
            'Fn::Sub': SubEvaluator(ref_evaluator, get_att_evaluator, self),
            'Fn::ImportValue': ImportValueEvaluator(self, account_config.region),
            'Fn::Split': SplitEvaluator(self),
            'Fn::Join': JoinEvaluator(self),
            'Fn::Select': SelectEvaluator(self),
            'Fn::FindInMap':  FindInMapEvaluator(mappings, self)
        }

    def eval_with_validation(self, value: Any, expected_schema: Any, resource_properties_to_eval=None, path: str = None, visited_values=None):
        value = self.eval(value, resource_properties_to_eval, visited_values)
        if value is not None:
            validate_schema(value, expected_schema, path)

        return value

    @prune_references_to_no_value
    @evaluate_dynamic_references
    def eval(self, value: Any, resource_properties_to_eval=None, visited_values=None):
        """ Evaluates the value of a CloudFormation key/value pair by evaluating intrinsic functions and pseudo
            parameters in the template and returns the evaluated value.

            value: the value of a key/value pair in a CloudFormation template
            resource_properties_to_eval: only evaluate these properties for a resource.  This limits the scope of intrinsic
                function support to be only those functions that we'd expect to exist in an IAM policy
                (e.g. Fn::GetAZs is not likely to appear)
            visited_values: tracks visited values to detect circular references in a CloudFormation template
        """
        visited_values = [] if visited_values is None else visited_values

        if isinstance(value, list):
            # when passing the list of visited values to child list items, we create a separate copy for each so
            # that they don't share the same visited references
            return [self.eval(item, copy.deepcopy(visited_values)) for item in value]

        elif isinstance(value, CfnObject):
            # intrinsic functions must be the only key, so we only need to look at the first key
            first_key = next(iter(value), None)
            evaluator = self.evaluators.get(first_key)
            if evaluator is not None:
                intrinsic_function_value = value.get(first_key)
                return evaluator.evaluate(intrinsic_function_value, visited_values=visited_values)

            # if it's not an intrinsic function, recursively traverse through child nodes
            for key in value.keys():
                # we only want to evaluate resource properties that we care about for IAM policies
                # if we were to evaluate all resource properties, the scope of supported intrinsic functions and
                # CFN parameters that would need to be passed would be larger.  This will ignore properties that we
                # don't care about.
                if value.parent == 'Properties':
                    if resource_properties_to_eval is not None and \
                            key not in resource_properties_to_eval:
                        continue

                # when passing the list of visited values to child evals, create a separate copy for each so
                # that they don't share the same visited references
                copy_of_visited_values = copy.deepcopy(visited_values)
                value[key] = self.eval(value[key], resource_properties_to_eval, copy_of_visited_values)

            return dict(value)

        else:
            return value
